from __future__ import annotations

from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass, field

import numpy as np
from typing_extensions import List, TYPE_CHECKING, Union, Optional, Dict, Any, Self

from semantic_digital_twin.world_description.geometry import transformation_from_json
from .degree_of_freedom import DegreeOfFreedom
from .world_entity import CollisionCheckingConfig, Connection, KinematicStructureEntity
from .. import spatial_types as cas
from ..datastructures.prefixed_name import PrefixedName
from ..datastructures.types import NpMatrix4x4
from ..spatial_types.derivatives import DerivativeMap

if TYPE_CHECKING:
    from ..world import World


class HasUpdateState(ABC):
    """
    Mixin class for connections that need state updated which are not trivial integrations.
    Typically needed for connections that use active and passive degrees of freedom.
    Look at OmniDrive for an example usage.
    """

    @abstractmethod
    def update_state(self, dt: float) -> None:
        """
        Allows the connection to update the state of its dofs.
        An integration update for active dofs will have happened before this method is called.
        Write directly into self._world.state, but don't touch dofs that don't belong to this connection.
        :param dt: Time passed since last update.
        """
        pass


@dataclass
class FixedConnection(Connection):
    """
    Has 0 degrees of freedom.
    """

    def __hash__(self):
        return hash((self.parent, self.child))


@dataclass
class ActiveConnection(Connection):
    """
    Has one or more degrees of freedom that can be actively controlled, e.g., robot joints.
    """

    frozen_for_collision_avoidance: bool = False
    """
    Should be treated as fixed for collision avoidance.
    Common example are gripper joints, you generally don't want to avoid collisions by closing the fingers, 
    but by moving the whole hand away.
    """

    def to_json(self) -> Dict[str, Any]:
        result = super().to_json()
        result["frozen_for_collision_avoidance"] = self.frozen_for_collision_avoidance
        return result

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(
            name=PrefixedName.from_json(data["name"]),
            parent=KinematicStructureEntity.from_json(data["parent"]),
            child=KinematicStructureEntity.from_json(data["child"]),
            parent_T_connection_expression=transformation_from_json(
                data["parent_T_connection_expression"]
            ),
            frozen_for_collision_avoidance=data["frozen_for_collision_avoidance"],
        )

    @property
    def has_hardware_interface(self) -> bool:
        """
        Whether this connection is linked to a controller and can therefore respond to control commands.

        E.g. the caster wheels of a PR2 are active, because they have a DOF, but they are not directly controlled.
        Instead a the omni drive connection is directly controlled and a low level controller translates these commands
        to commands for the caster wheels.

        A door hinge is also active but cannot be controlled.
        """
        return any(dof.has_hardware_interface for dof in self.dofs)

    @has_hardware_interface.setter
    def has_hardware_interface(self, value: bool) -> None:
        for dof in self.dofs:
            dof.has_hardware_interface = value

    @property
    def active_dofs(self) -> List[DegreeOfFreedom]:
        return []

    def set_static_collision_config_for_direct_child_bodies(
        self, collision_config: CollisionCheckingConfig
    ):
        for child_body in self._world.get_direct_child_bodies_with_collision(self):
            if not child_body.get_collision_config().disabled:
                child_body.set_static_collision_config(collision_config)


@dataclass
class PassiveConnection(Connection):
    """
    Has one or more degrees of freedom that cannot be actively controlled.
    Useful if a transformation is only tracked, e.g., the robot's localization.
    """

    @property
    def passive_dofs(self) -> List[DegreeOfFreedom]:
        return []


@dataclass
class ActiveConnection1DOF(ActiveConnection, ABC):
    """
    Superclass for active connections with 1 degree of freedom.
    """

    axis: cas.Vector3 = field(kw_only=True)
    """
    Connection moves along this axis, should be a unit vector.
    The axis is defined relative to the local reference frame of the parent KinematicStructureEntity.
    """

    multiplier: float = 1.0
    """
    Movement along the axis is multiplied by this value. Useful if Connections share DoFs.
    """

    offset: float = 0.0
    """
    Movement along the axis is offset by this value. Useful if Connections share DoFs.
    """

    dof_name: PrefixedName = field(kw_only=True)
    """
    Name of a Degree of freedom to control movement along the axis.
    """

    def to_json(self) -> Dict[str, Any]:
        result = super().to_json()
        result["axis"] = self.axis.to_np().tolist()
        result["multiplier"] = self.multiplier
        result["offset"] = self.offset
        result["dof_name"] = self.dof_name.to_json()
        return result

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(
            name=PrefixedName.from_json(data["name"]),
            parent=KinematicStructureEntity.from_json(data["parent"]),
            child=KinematicStructureEntity.from_json(data["child"]),
            parent_T_connection_expression=transformation_from_json(
                data["parent_T_connection_expression"]
            ),
            frozen_for_collision_avoidance=data["frozen_for_collision_avoidance"],
            axis=cas.Vector3.from_iterable(data["axis"]),
            multiplier=data["multiplier"],
            offset=data["offset"],
            dof_name=PrefixedName.from_json(data["dof_name"]),
        )

    @classmethod
    def create_with_dofs(
        cls,
        world: World,
        parent: KinematicStructureEntity,
        child: KinematicStructureEntity,
        axis: cas.Vector3,
        name: Optional[PrefixedName] = None,
        multiplier: float = 1.0,
        offset: float = 0.0,
    ) -> Self:
        """
        Creates and returns an instance of the class with associated degrees of freedom
        (DOFs) based on the specified parameters. This method facilitates initializing
        a kinematic relationship between a parent and a child entity, augmented by
        an axis representation and configurable properties such as multiplier and offset.

        :param world: The motion world in which to add the degree of freedom.
        :param parent: The parent kinematic structure entity.
        :param child: The child kinematic structure entity.
        :param axis: The axis vector defining the joint relation.
        :param name: Optional specific name for the DOF entity. If not provided, a
                     default name is generated based on the parent and child.
        :param multiplier: A scaling factor applied to the DOF's motion. Defaults to 1.0.
        :param offset: A constant offset value applied to the DOF's motion. Defaults to 0.0.
        :return: An instance of the class representing the defined relationship with
                 its DOF added to the world.
        """
        name = name or cls._generate_default_name(parent=parent, child=child)
        dof = DegreeOfFreedom(name=PrefixedName("dof", str(name)))
        world.add_degree_of_freedom(dof)
        return cls(
            parent=parent,
            child=child,
            axis=axis,
            multiplier=multiplier,
            offset=offset,
            dof_name=dof.name,
        )

    def add_to_world(self, world: World):
        super().add_to_world(world)
        if self.multiplier is None:
            self.multiplier = 1
        else:
            self.multiplier = self.multiplier
        if self.offset is None:
            self.offset = 0
        else:
            self.offset = self.offset
        self.axis = self.axis

    @property
    def dof(self) -> DegreeOfFreedom:
        """
        A reference to the Degree of Freedom associated with this connection.
        .. warning:: WITH multiplier and offset applied.
        """
        result = deepcopy(self.raw_dof)
        result.symbols = result.symbols * self.multiplier
        if self.multiplier < 0:
            # if multiplier is negative, we need to swap the limits
            result.lower_limits, result.upper_limits = (
                result.upper_limits,
                result.lower_limits,
            )
        result.lower_limits = result.lower_limits * self.multiplier
        result.upper_limits = result.upper_limits * self.multiplier

        result.symbols.position += self.offset
        if result.lower_limits.position is not None:
            result.lower_limits.position = result.lower_limits.position + self.offset
        if result.upper_limits.position is not None:
            result.upper_limits.position = result.upper_limits.position + self.offset
        return result

    @property
    def raw_dof(self) -> DegreeOfFreedom:
        """
        A reference to the Degree of Freedom associated with this connection.
        .. warning:: WITHOUT multiplier and offset applied.
        """
        return self._world.get_degree_of_freedom_by_name(self.dof_name)

    @property
    def active_dofs(self) -> List[DegreeOfFreedom]:
        return [self.raw_dof]

    def __hash__(self):
        return hash((self.parent, self.child))

    @property
    def position(self) -> float:
        return self._world.state[self.dof.name].position * self.multiplier + self.offset

    @position.setter
    def position(self, value: float) -> None:
        self._world.state[self.dof.name].position = (
            value - self.offset
        ) / self.multiplier
        self._world.notify_state_change()

    @property
    def velocity(self) -> float:
        return self._world.state[self.dof.name].velocity * self.multiplier

    @velocity.setter
    def velocity(self, value: float) -> None:
        self._world.state[self.dof.name].velocity = value / self.multiplier
        self._world.notify_state_change()

    @property
    def acceleration(self) -> float:
        return self._world.state[self.dof.name].acceleration * self.multiplier

    @acceleration.setter
    def acceleration(self, value: float) -> None:
        self._world.state[self.dof.name].acceleration = value / self.multiplier
        self._world.notify_state_change()

    @property
    def jerk(self) -> float:
        return self._world.state[self.dof.name].jerk * self.multiplier

    @jerk.setter
    def jerk(self, value: float) -> None:
        self._world.state[self.dof.name].jerk = value / self.multiplier
        self._world.notify_state_change()


@dataclass
class PrismaticConnection(ActiveConnection1DOF):
    """
    Allows translation along an axis.
    """

    def add_to_world(self, world: World):
        super().add_to_world(world)

        translation_axis = self.axis * self.dof.symbols.position
        self.connection_T_child_expression = cas.TransformationMatrix.from_xyz_rpy(
            x=translation_axis[0],
            y=translation_axis[1],
            z=translation_axis[2],
            child_frame=self.child,
        )

    def __hash__(self):
        return hash((self.parent, self.child))


@dataclass
class RevoluteConnection(ActiveConnection1DOF):
    """
    Allows rotation about an axis.
    """

    def add_to_world(self, world: World):
        super().add_to_world(world)

        self.connection_T_child_expression = (
            cas.TransformationMatrix.from_xyz_axis_angle(
                axis=self.axis,
                angle=self.dof.symbols.position,
                child_frame=self.child,
            )
        )

    def __hash__(self):
        return hash((self.parent, self.child))


@dataclass
class Connection6DoF(PassiveConnection):
    """
    Has full 6 degrees of freedom, that cannot be actively controlled.
    Useful for synchronizing with transformations from external providers.
    """

    x_name: PrefixedName = field(kw_only=True)
    """
    Displacement of child KinematicStructureEntity with respect to parent KinematicStructureEntity along the x-axis.
    """
    y_name: PrefixedName = field(kw_only=True)
    """
    Displacement of child KinematicStructureEntity with respect to parent KinematicStructureEntity along the y-axis.
    """
    z_name: PrefixedName = field(kw_only=True)
    """
    Displacement of child KinematicStructureEntity with respect to parent KinematicStructureEntity along the z-axis.
    """

    qx_name: PrefixedName = field(kw_only=True)
    qy_name: PrefixedName = field(kw_only=True)
    qz_name: PrefixedName = field(kw_only=True)
    qw_name: PrefixedName = field(kw_only=True)
    """
    Rotation of child KinematicStructureEntity with respect to parent KinematicStructureEntity represented as a quaternion.
    """

    def to_json(self) -> Dict[str, Any]:
        result = super().to_json()
        result["x_name"] = self.x_name.to_json()
        result["y_name"] = self.y_name.to_json()
        result["z_name"] = self.z_name.to_json()
        result["qx_name"] = self.qx_name.to_json()
        result["qy_name"] = self.qy_name.to_json()
        result["qz_name"] = self.qz_name.to_json()
        result["qw_name"] = self.qw_name.to_json()
        return result

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(
            name=PrefixedName.from_json(data["name"]),
            parent=KinematicStructureEntity.from_json(data["parent"]),
            child=KinematicStructureEntity.from_json(data["child"]),
            parent_T_connection_expression=transformation_from_json(
                data["parent_T_connection_expression"]
            ),
            x_name=PrefixedName.from_json(data["x_name"]),
            y_name=PrefixedName.from_json(data["y_name"]),
            z_name=PrefixedName.from_json(data["z_name"]),
            qx_name=PrefixedName.from_json(data["qx_name"]),
            qy_name=PrefixedName.from_json(data["qy_name"]),
            qz_name=PrefixedName.from_json(data["qz_name"]),
            qw_name=PrefixedName.from_json(data["qw_name"]),
        )

    @property
    def x(self) -> DegreeOfFreedom:
        return self._world.get_degree_of_freedom_by_name(self.x_name)

    @property
    def y(self) -> DegreeOfFreedom:
        return self._world.get_degree_of_freedom_by_name(self.y_name)

    @property
    def z(self) -> DegreeOfFreedom:
        return self._world.get_degree_of_freedom_by_name(self.z_name)

    @property
    def qx(self) -> DegreeOfFreedom:
        return self._world.get_degree_of_freedom_by_name(self.qx_name)

    @property
    def qy(self) -> DegreeOfFreedom:
        return self._world.get_degree_of_freedom_by_name(self.qy_name)

    @property
    def qz(self) -> DegreeOfFreedom:
        return self._world.get_degree_of_freedom_by_name(self.qz_name)

    @property
    def qw(self) -> DegreeOfFreedom:
        return self._world.get_degree_of_freedom_by_name(self.qw_name)

    def __hash__(self):
        return hash(self.name)

    def add_to_world(self, world: World):
        super().add_to_world(world)
        parent_P_child = cas.Point3(
            x_init=self.x.symbols.position,
            y_init=self.y.symbols.position,
            z_init=self.z.symbols.position,
        )
        parent_R_child = cas.Quaternion(
            x_init=self.qx.symbols.position,
            y_init=self.qy.symbols.position,
            z_init=self.qz.symbols.position,
            w_init=self.qw.symbols.position,
        ).to_rotation_matrix()
        self.connection_T_child_expression = (
            cas.TransformationMatrix.from_point_rotation_matrix(
                point=parent_P_child,
                rotation_matrix=parent_R_child,
                child_frame=self.child,
            )
        )

    @classmethod
    def create_with_dofs(
        cls,
        world: World,
        parent: KinematicStructureEntity,
        child: KinematicStructureEntity,
        name: Optional[PrefixedName] = None,
        parent_T_connection_expression: Optional[cas.TransformationMatrix] = None,
    ) -> Self:
        """
        Creates an instance of the class with automatically generated degrees of freedom (DoFs)
        for the provided parent and child kinematic entities within the specified world.

        This method initializes and adds the required degrees of freedom to the world,
        and sets their properties accordingly. It generates a name for the connection if
        none is provided, and ensures valid initial state for relevant degrees of freedom.

        :param world: The World object where the degrees of freedom are added and modified.
        :param parent: The KinematicStructureEntity serving as the parent.
        :param child: The KinematicStructureEntity serving as the child.
        :param name: An optional PrefixedName for the connection. If None, it will be
                     auto-generated based on the parent and child names.
        :param parent_T_connection_expression: Optional transformation matrix specifying
                                               the connection relationship between parent
                                               and child entities.
        :return: A new instance of the class representing the parent-child connection with
                 automatically defined degrees of freedom.
        """
        name = name or cls._generate_default_name(parent=parent, child=child)

        with world.modify_world():
            stringified_name = str(name)
            x = DegreeOfFreedom(name=PrefixedName("x", stringified_name))
            world.add_degree_of_freedom(x)
            y = DegreeOfFreedom(name=PrefixedName("y", stringified_name))
            world.add_degree_of_freedom(y)
            z = DegreeOfFreedom(name=PrefixedName("z", stringified_name))
            world.add_degree_of_freedom(z)
            qx = DegreeOfFreedom(name=PrefixedName("qx", stringified_name))
            world.add_degree_of_freedom(qx)
            qy = DegreeOfFreedom(name=PrefixedName("qy", stringified_name))
            world.add_degree_of_freedom(qy)
            qz = DegreeOfFreedom(name=PrefixedName("qz", stringified_name))
            world.add_degree_of_freedom(qz)
            qw = DegreeOfFreedom(name=PrefixedName("qw", stringified_name))
            world.add_degree_of_freedom(qw)
            world.state[qw.name].position = 1.0

        return cls(
            parent=parent,
            child=child,
            parent_T_connection_expression=parent_T_connection_expression,
            name=name,
            x_name=x.name,
            y_name=y.name,
            z_name=z.name,
            qx_name=qx.name,
            qy_name=qy.name,
            qz_name=qz.name,
            qw_name=qw.name,
        )

    @property
    def passive_dofs(self) -> List[DegreeOfFreedom]:
        return [self.x, self.y, self.z, self.qx, self.qy, self.qz, self.qw]

    @property
    def origin(self) -> cas.TransformationMatrix:
        return super().origin

    @origin.setter
    def origin(
        self, transformation: Union[NpMatrix4x4, cas.TransformationMatrix]
    ) -> None:
        if not isinstance(transformation, cas.TransformationMatrix):
            transformation = cas.TransformationMatrix(data=transformation)
        position = transformation.to_position().to_np()
        orientation = transformation.to_rotation_matrix().to_quaternion().to_np()
        self._world.state[self.x.name].position = position[0]
        self._world.state[self.y.name].position = position[1]
        self._world.state[self.z.name].position = position[2]
        self._world.state[self.qx.name].position = orientation[0]
        self._world.state[self.qy.name].position = orientation[1]
        self._world.state[self.qz.name].position = orientation[2]
        self._world.state[self.qw.name].position = orientation[3]
        self._world.notify_state_change()


@dataclass
class OmniDrive(ActiveConnection, PassiveConnection, HasUpdateState):
    """
    A connection describing an omnidirectional drive.
    It can rotate about its z-axis and drive on the x-y plane simultaneously.
    - x/y: Passive dofs describing the measured odometry with respect to parent frame.
        We assume that the robot can't fly, and we can't measure its z-axis position, so z=0.
        The odometry sensors typically provide velocity measurements with respect to the child frame,
        therefore the velocity values of x/y must stay 0.
    - x_vel/y_vel: The measured and commanded velocity is represented with respect to the child frame with these
        active dofs. It must be ensured that their position values stay 0.
    - roll/pitch: Some robots, like the PR2, have sensors to measure pitch and roll using an IMU,
        we therefore have passive dofs for them.
    - yaw: Since the robot can only rotate about its z-axis, we don't need different dofs for position and velocity of yaw.
        They are combined into one active dof.
    """

    # passive dofs
    x_name: PrefixedName = field(kw_only=True)
    y_name: PrefixedName = field(kw_only=True)
    roll_name: PrefixedName = field(kw_only=True)
    pitch_name: PrefixedName = field(kw_only=True)

    # active dofs
    yaw_name: PrefixedName = field(kw_only=True)
    x_velocity_name: PrefixedName = field(kw_only=True)
    y_velocity_name: PrefixedName = field(kw_only=True)

    def to_json(self) -> Dict[str, Any]:
        result = super().to_json()
        result["x_name"] = self.x_name.to_json()
        result["y_name"] = self.y_name.to_json()
        result["roll_name"] = self.roll_name.to_json()
        result["pitch_name"] = self.pitch_name.to_json()
        result["yaw_name"] = self.yaw_name.to_json()
        result["x_velocity_name"] = self.x_velocity_name.to_json()
        result["y_velocity_name"] = self.y_velocity_name.to_json()
        return result

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(
            name=PrefixedName.from_json(data["name"]),
            parent=KinematicStructureEntity.from_json(data["parent"]),
            child=KinematicStructureEntity.from_json(data["child"]),
            parent_T_connection_expression=transformation_from_json(
                data["parent_T_connection_expression"]
            ),
            x_name=PrefixedName.from_json(data["x_name"]),
            y_name=PrefixedName.from_json(data["y_name"]),
            roll_name=PrefixedName.from_json(data["roll_name"]),
            pitch_name=PrefixedName.from_json(data["pitch_name"]),
            yaw_name=PrefixedName.from_json(data["yaw_name"]),
            x_velocity_name=PrefixedName.from_json(data["x_velocity_name"]),
            y_velocity_name=PrefixedName.from_json(data["y_velocity_name"]),
        )

    @property
    def x(self) -> DegreeOfFreedom:
        return self._world.get_degree_of_freedom_by_name(self.x_name)

    @property
    def y(self) -> DegreeOfFreedom:
        return self._world.get_degree_of_freedom_by_name(self.y_name)

    @property
    def roll(self) -> DegreeOfFreedom:
        return self._world.get_degree_of_freedom_by_name(self.roll_name)

    @property
    def pitch(self) -> DegreeOfFreedom:
        return self._world.get_degree_of_freedom_by_name(self.pitch_name)

    @property
    def yaw(self) -> DegreeOfFreedom:
        return self._world.get_degree_of_freedom_by_name(self.yaw_name)

    @property
    def x_velocity(self) -> DegreeOfFreedom:
        return self._world.get_degree_of_freedom_by_name(self.x_velocity_name)

    @property
    def y_velocity(self) -> DegreeOfFreedom:
        return self._world.get_degree_of_freedom_by_name(self.y_velocity_name)

    def add_to_world(self, world: World):
        super().add_to_world(world)
        odom_T_bf = cas.TransformationMatrix.from_xyz_rpy(
            x=self.x.symbols.position,
            y=self.y.symbols.position,
            yaw=self.yaw.symbols.position,
        )
        bf_T_bf_vel = cas.TransformationMatrix.from_xyz_rpy(
            x=self.x_velocity.symbols.position, y=self.y_velocity.symbols.position
        )
        bf_vel_T_bf = cas.TransformationMatrix.from_xyz_rpy(
            x=0,
            y=0,
            z=0,
            roll=self.roll.symbols.position,
            pitch=self.pitch.symbols.position,
            yaw=0,
        )
        self.connection_T_child_expression = odom_T_bf @ bf_T_bf_vel @ bf_vel_T_bf
        self.connection_T_child_expression.child_frame = self.child

    @classmethod
    def create_with_dofs(
        cls,
        world: World,
        parent: KinematicStructureEntity,
        child: KinematicStructureEntity,
        name: Optional[PrefixedName] = None,
        parent_T_connection_expression: Optional[cas.TransformationMatrix] = None,
        translation_velocity_limits: float = 0.6,
        rotation_velocity_limits: float = 0.5,
    ) -> Self:
        """
        Creates an instance of the class with automatically generated degrees of freedom
        (DOFs) for translation on the x and y axes, rotation along roll, pitch, and yaw
        axes, and velocity limits for translation and rotation.

        This method modifies the provided world to add all required degrees of freedom
        and their limits, based on the provided settings. Names for the degrees of
        freedom are auto-generated using the stringified version of the provided name
        or its default setting.

        :param world: The world where the configuration is being applied, and degrees of freedom are added.
        :param parent: The parent kinematic structure entity.
        :param child: The child kinematic structure entity.
        :param name: Name of the connection. If None, it will be auto-generated.
        :param parent_T_connection_expression: Transformation matrix representing the
            relative position/orientation of the child to the parent. Default is Identity.
        :param translation_velocity_limits: The velocity limit applied to the
            translation degrees of freedom (default is 0.6).
        :param rotation_velocity_limits: The velocity limit applied to the rotation
            degrees of freedom (default is 0.5).
        :return: An instance of the class with the auto-generated DOFs incorporated.
        """
        name = name or cls._generate_default_name(parent=parent, child=child)
        with world.modify_world():
            stringified_name = str(name)
            lower_translation_limits = DerivativeMap()
            lower_translation_limits.velocity = -translation_velocity_limits
            upper_translation_limits = DerivativeMap()
            upper_translation_limits.velocity = translation_velocity_limits
            lower_rotation_limits = DerivativeMap()
            lower_rotation_limits.velocity = -rotation_velocity_limits
            upper_rotation_limits = DerivativeMap()
            upper_rotation_limits.velocity = rotation_velocity_limits

            with world.modify_world():
                x = DegreeOfFreedom(name=PrefixedName("x", stringified_name))
                world.add_degree_of_freedom(x)
                y = DegreeOfFreedom(name=PrefixedName("y", stringified_name))
                world.add_degree_of_freedom(y)
                roll = DegreeOfFreedom(name=PrefixedName("roll", stringified_name))
                world.add_degree_of_freedom(roll)
                pitch = DegreeOfFreedom(name=PrefixedName("pitch", stringified_name))
                world.add_degree_of_freedom(pitch)
                yaw = DegreeOfFreedom(
                    name=PrefixedName("yaw", stringified_name),
                    lower_limits=lower_rotation_limits,
                    upper_limits=upper_rotation_limits,
                )
                world.add_degree_of_freedom(yaw)

                x_vel = DegreeOfFreedom(
                    name=PrefixedName("x_vel", stringified_name),
                    lower_limits=lower_translation_limits,
                    upper_limits=upper_translation_limits,
                )
                world.add_degree_of_freedom(x_vel)
                y_vel = DegreeOfFreedom(
                    name=PrefixedName("y_vel", stringified_name),
                    lower_limits=lower_translation_limits,
                    upper_limits=upper_translation_limits,
                )
                world.add_degree_of_freedom(y_vel)

        return cls(
            parent=parent,
            child=child,
            parent_T_connection_expression=parent_T_connection_expression,
            name=name,
            x_name=x.name,
            y_name=y.name,
            roll_name=roll.name,
            pitch_name=pitch.name,
            yaw_name=yaw.name,
            x_velocity_name=x_vel.name,
            y_velocity_name=y_vel.name,
        )

    @property
    def active_dofs(self) -> List[DegreeOfFreedom]:
        return [self.x_velocity, self.y_velocity, self.yaw]

    @property
    def passive_dofs(self) -> List[DegreeOfFreedom]:
        return [self.x, self.y, self.roll, self.pitch]

    @property
    def dofs(self) -> List[DegreeOfFreedom]:
        return self.active_dofs + self.passive_dofs

    def update_state(self, dt: float) -> None:
        state = self._world.state
        state[self.x_velocity.name].position = 0
        state[self.y_velocity.name].position = 0

        x_vel = state[self.x_velocity.name].velocity
        y_vel = state[self.y_velocity.name].velocity
        delta = state[self.yaw.name].position
        x_velocity = np.cos(delta) * x_vel - np.sin(delta) * y_vel
        state[self.x.name].position += x_velocity * dt
        y_velocity = np.sin(delta) * x_vel + np.cos(delta) * y_vel
        state[self.y.name].position += y_velocity * dt

    @property
    def origin(self) -> cas.TransformationMatrix:
        return super().origin

    @origin.setter
    def origin(
        self, transformation: Union[NpMatrix4x4, cas.TransformationMatrix]
    ) -> None:
        if isinstance(transformation, np.ndarray):
            transformation = cas.TransformationMatrix(data=transformation)
        position = transformation.to_position()
        roll, pitch, yaw = transformation.to_rotation_matrix().to_rpy()
        assert (
            position.z.to_np() == 0.0
        ), "OmniDrive only supports planar movement in the XY plane, z must be 0"
        assert (
            roll.to_np() == 0.0
        ), "OmniDrive only supports planar movement in the XY plane, roll must be 0"
        assert (
            pitch.to_np() == 0.0
        ), "OmniDrive only supports planar movement in the XY plane, pitch must be 0"
        self._world.state[self.x.name].position = position.x.to_np()
        self._world.state[self.y.name].position = position.y.to_np()
        self._world.state[self.yaw.name].position = yaw.to_np()
        self._world.notify_state_change()

    def get_free_variable_names(self) -> List[PrefixedName]:
        return [self.x.name, self.y.name, self.yaw.name]

    def __hash__(self):
        return hash(self.name)

    @property
    def has_hardware_interface(self) -> bool:
        return self.x_velocity.has_hardware_interface

    @has_hardware_interface.setter
    def has_hardware_interface(self, value: bool) -> None:
        self.x_velocity.has_hardware_interface = value
        self.y_velocity.has_hardware_interface = value
        self.yaw.has_hardware_interface = value



from semantic_digital_twin.datastructures.prefixed_name import PrefixedName
from semantic_digital_twin.spatial_types.spatial_types import TransformationMatrix
from semantic_digital_twin.world import World
from semantic_digital_twin.world_description.connections import FixedConnection
from semantic_digital_twin.world_description.geometry import Sphere
from semantic_digital_twin.world_description.shape_collection import ShapeCollection
from semantic_digital_twin.world_description.world_entity import Body


def test_post_init_transformation():
    w = World()
    root = Body(name=PrefixedName("root"))
    b1 = Body(name=PrefixedName("b1"))

    with w.modify_world():
        w.add_connection(
            FixedConnection(
                parent=root,
                child=b1,
                parent_T_connection_expression=TransformationMatrix.from_xyz_rpy(
                    x=1, reference_frame=root
                ),
            )
        )

    shape = Sphere(
        radius=1, origin=TransformationMatrix.from_xyz_rpy(x=3, reference_frame=root)
    )
    shape_collection = ShapeCollection(
        shapes=[shape],
        reference_frame=b1,
    )
    shape_collection.transform_all_shapes_to_own_frame()
    assert shape.origin.reference_frame == b1
    assert shape.origin.to_position().x == 2.0

    shape = Sphere(
        radius=1, origin=TransformationMatrix.from_xyz_rpy(x=3, reference_frame=root)
    )

    shape_collection = ShapeCollection(reference_frame=b1)
    shape_collection.append(shape)
    shape_collection.transform_all_shapes_to_own_frame()
    assert shape.origin.reference_frame == b1
    assert shape.origin.to_position().x == 2.0



from __future__ import annotations

import abc
from dataclasses import dataclass, field
from itertools import chain
from typing_extensions import Tuple, Set, List, Optional, Iterable, TYPE_CHECKING

import numpy as np

from ..world_description.connections import ActiveConnection
from ..world_description.world_entity import Body

if TYPE_CHECKING:
    from ..world import World


@dataclass
class CollisionCheck:
    body_a: Body
    """
    First body in the collision check.
    """
    body_b: Body
    """
    Second body in the collision check.
    """
    distance: float
    """
    Minimum distance to check for collisions.
    """
    _world: World
    """
    The world context for validation and sorting.
    """

    def __post_init__(self):
        self.body_a, self.body_b = self.sort_bodies(self.body_a, self.body_b)

    def __hash__(self):
        return hash((self.body_a, self.body_b))

    def __eq__(self, other: CollisionCheck):
        return self.body_a == other.body_a and self.body_b == other.body_b

    def bodies(self) -> Tuple[Body, Body]:
        return self.body_a, self.body_b

    def _validate(self) -> None:
        """Validates the collision check parameters."""
        if self.distance <= 0:
            raise ValueError(f"Distance must be positive, got {self.distance}")

        if self.body_a == self.body_b:
            raise ValueError(
                f'Cannot create collision check between the same body "{self.body_a.name}"'
            )

        if not self.body_a.has_collision():
            raise ValueError(f"Body {self.body_a.name} has no collision geometry")

        if not self.body_b.has_collision():
            raise ValueError(f"Body {self.body_b.name} has no collision geometry")

        if self.body_a not in self._world.bodies_with_enabled_collision:
            raise ValueError(
                f"Body {self.body_a.name} is not in list of bodies with collisions"
            )

        if self.body_b not in self._world.bodies_with_enabled_collision:
            raise ValueError(
                f"Body {self.body_b.name} is not in list of bodies with collisions"
            )

        root_chain, tip_chain = self._world.compute_split_chain_of_connections(
            self.body_a, self.body_b
        )
        if all(
            not isinstance(c, ActiveConnection) for c in chain(root_chain, tip_chain)
        ):
            raise ValueError(
                f"Relative pose between {self.body_a.name} and {self.body_b.name} is fixed"
            )

    @classmethod
    def create_and_validate(
        cls, body_a: Body, body_b: Body, distance: float, world: World
    ) -> CollisionCheck:
        """
        Creates a collision check with additional world-context validation.
        Returns None if the check should be skipped (e.g., bodies are linked).
        """
        collision_check = cls(
            body_a=body_a, body_b=body_b, distance=distance, _world=world
        )
        collision_check._validate()
        return collision_check

    def sort_bodies(self, body_a: Body, body_b: Body) -> Tuple[Body, Body]:
        """
        Sort both bodies in a consistent manner, needed to avoid checking B with A, when A with B is already checked.
        """
        if body_a.name > body_b.name:
            body_a, body_b = body_b, body_a
        is_body_a_controlled = self._world.is_body_controlled(body_a)
        is_body_b_controlled = self._world.is_body_controlled(body_b)
        if not is_body_a_controlled and is_body_b_controlled:
            body_a, body_b = body_b, body_a
        return body_a, body_b


@dataclass
class Collision:
    contact_distance: float
    body_a: Body = field(default=None)
    """
    First body in the collision.
    """
    body_b: Body = field(default=None)
    """
    Second body in the collision.
    """

    map_P_pa: np.ndarray = field(default=None)
    """
    Contact point on body A with respect to the world frame.
    """
    map_P_pb: np.ndarray = field(default=None)
    """
    Contact point on body B with respect to the world frame.
    """
    map_V_n_input: np.ndarray = field(default=None)
    """
    Contact normal with respect to the world frame.
    """
    a_P_pa: np.ndarray = field(default=None)
    """
    Contact point on body A with respect to the link a frame. 
    """
    b_P_pb: np.ndarray = field(default=None)
    """
    Contact point on body B with respect to the link b frame.
    """

    def __str__(self):
        return f"{self.body_a}|<->|{self.body_b}: {self.contact_distance}"

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash((self.body_a, self.body_b))

    def __eq__(self, other: CollisionCheck):
        return self.body_a == other.body_a and self.body_b == other.body_b


@dataclass
class CollisionDetector(abc.ABC):
    """
    Abstract class for collision detectors.
    """

    _world: World

    @abc.abstractmethod
    def sync_world_model(self) -> None:
        """
        Synchronize the collision checker with the current world model
        """

    @abc.abstractmethod
    def sync_world_state(self) -> None:
        """
        Synchronize the collision checker with the current world state
        """

    @abc.abstractmethod
    def check_collisions(
        self, collision_matrix: Optional[Iterable[CollisionCheck]] = None
    ) -> List[Collision]:
        """
        Computes the collisions for all checks in the collision matrix.
        If collision_matrix is None, checks all collisions.
        :param collision_matrix:
        :return: A list of detected collisions.
        """

    @abc.abstractmethod
    def reset_cache(self):
        """
        Reset any caches the collision checker may have.
        """



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum

import fbxloader
import numpy as np
import trimesh
from fbxloader import Object3D, Mesh as FBXMesh, Scene

from ..adapters.mesh import MeshParser
from ..datastructures.prefixed_name import PrefixedName
from ..spatial_types import (
    TransformationMatrix,
)
from ..world import World
from ..world_description.connections import FixedConnection
from ..world_description.geometry import TriangleMesh
from ..world_description.shape_collection import ShapeCollection
from ..world_description.world_entity import Body


@dataclass
class CoordinateAxis(Enum):
    """
    Enum for coordinate axes with direction.
    The value is a tuple of (axis_index, sign), where axis_index is 0 for X, 1 for Y, and 2 for Z,
    and sign is 1 for positive direction and -1 for negative direction.
    """

    POSITIVE_X = (0, 1)
    NEGATIVE_X = (0, -1)
    POSITIVE_Y = (1, 1)
    NEGATIVE_Y = (1, -1)
    POSITIVE_Z = (2, 1)
    NEGATIVE_Z = (2, -1)

    @classmethod
    def from_fbx(cls, axis_value: int, sign: int) -> CoordinateAxis:
        """Map FBX axis index + sign into a CoordinateAxis enum."""
        assert axis_value in (0, 1, 2), "axis_value must be 0, 1, or 2"
        assert sign in (1, -1), "sign must be 1 or -1"
        for member in cls:
            if member.value[0] == axis_value and member.value[1] == sign:
                return member

    def to_vector(self) -> np.ndarray:
        """Convert the CoordinateAxis to a 3D unit vector."""
        idx, sgn = self.value
        v = np.zeros(3)
        v[idx] = float(sgn)
        return v


@dataclass
class FBXGlobalSettings:
    """
    Class to handle FBX global settings, particularly the coordinate system.
    This class extracts the up, front, and coordinate axes from the FBX file and provides
    a method to get the transformation matrix from FBX to Semantic Digital Twin coordinate system.
    """

    fbx_loader: fbxloader.FBXLoader
    """
    The FBXLoader instance containing the loaded FBX file.
    """

    up_axis: CoordinateAxis = field(init=False)
    """The up axis of the FBX file."""

    front_axis: CoordinateAxis = field(init=False)
    """The front axis of the FBX file."""

    coord_axis: CoordinateAxis = field(init=False)
    """The final, third axis of the FBX file, called the coordinate axis."""

    def __post_init__(self):
        fbx = self.fbx_loader

        self.up_axis = CoordinateAxis.from_fbx(
            fbx.fbxtree["GlobalSettings"]["UpAxis"]["value"],
            fbx.fbxtree["GlobalSettings"]["UpAxisSign"]["value"],
        )

        self.front_axis = CoordinateAxis.from_fbx(
            fbx.fbxtree["GlobalSettings"]["FrontAxis"]["value"],
            fbx.fbxtree["GlobalSettings"]["FrontAxisSign"]["value"],
        )

        self.coord_axis = CoordinateAxis.from_fbx(
            fbx.fbxtree["GlobalSettings"]["CoordAxis"]["value"],
            fbx.fbxtree["GlobalSettings"]["CoordAxisSign"]["value"],
        )

    def get_semantic_digital_twin_T_fbx(self) -> np.ndarray:
        """
        Get the transformation matrix from FBX to Semantic Digital Twin coordinate system.
        """
        sX = self.front_axis.to_vector()
        sY = self.coord_axis.to_vector()
        sZ = self.up_axis.to_vector()
        S = np.column_stack((sX, sY, sZ))

        R = S.T

        M = np.eye(4)
        M[:3, :3] = R
        return M


@dataclass
class FBXParser(MeshParser):
    """
    Adapter for FBX files.
    """

    @staticmethod
    def transform_vertices(
        fbx_vertices: np.ndarray, semantic_digital_twin_T_fbx: np.ndarray
    ) -> np.ndarray:
        """
        Transform vertices from FBX coordinate system to Semantic Digital Twin coordinate system.
        """
        assert (
            fbx_vertices.ndim == 2 and fbx_vertices.shape[1] == 3
        ), "vertices must be (N,3)"
        assert semantic_digital_twin_T_fbx.shape == (
            4,
            4,
        ), "semantic_digital_twin_T_fbx must be 4x4"

        fbx_vertices_h = np.hstack(
            [
                fbx_vertices,
                np.ones((fbx_vertices.shape[0], 1), dtype=fbx_vertices.dtype),
            ]
        )

        semantic_digital_twin_vertices = fbx_vertices_h @ np.linalg.inv(
            semantic_digital_twin_T_fbx
        )
        return semantic_digital_twin_vertices[:, :3]

    def parse(self) -> World:
        """
        Parse the FBX file, each object in the FBX file is converted to a body in the world and the meshes are loaded
        as TriangleMesh objects.

        :return: A World containing content of the FBX file.
        """
        fbx = fbxloader.FBXLoader(self.file_path)

        global_settings = FBXGlobalSettings(fbx)
        semantic_digital_twin_T_fbx = global_settings.get_semantic_digital_twin_T_fbx()

        world = World()

        with world.modify_world():
            for obj_id, obj in fbx.objects.items():
                # Create a body for each object in the FBX file
                if type(obj) is Object3D:
                    name = fbx.fbxtree["Objects"]["Model"][obj_id]["attrName"].split(
                        "\x00"
                    )[0]
                    meshes = []
                    for o in obj.children:
                        if isinstance(o, FBXMesh):
                            transformed_vertices = (
                                self.transform_vertices(
                                    o.vertices, semantic_digital_twin_T_fbx
                                )
                                / 100
                            )

                            t_mesh = TriangleMesh(
                                origin=TransformationMatrix(),
                                mesh=trimesh.Trimesh(
                                    vertices=transformed_vertices, faces=o.faces
                                ),
                            )

                            meshes.append(t_mesh)
                    body = Body(
                        name=PrefixedName(name),
                        collision=ShapeCollection(meshes),
                        visual=ShapeCollection(meshes),
                    )
                    world.add_body(body)

            for obj in fbx.objects.values():
                if type(obj) is Object3D:
                    name = fbx.fbxtree["Objects"]["Model"][obj.id]["attrName"].split(
                        "\x00"
                    )[0]
                    parent_name = (
                        fbx.fbxtree["Objects"]["Model"][obj.parent.id][
                            "attrName"
                        ].split("\x00")[0]
                        if type(obj.parent) is not Scene
                        else None
                    )
                    if not parent_name:
                        continue

                    obj_body = world.get_body_by_name(name)
                    parent_body = world.get_body_by_name(parent_name)

                    parent_T_child = TransformationMatrix(obj.matrix)

                    connection = FixedConnection(
                        parent=parent_body,
                        child=obj_body,
                        parent_T_connection_expression=parent_T_child,
                    )
                    world.add_connection(connection)

        return world



import json

from std_srvs.srv import Trigger

from semantic_digital_twin.adapters.ros.world_fetcher import (
    FetchWorldServer,
    fetch_world_from_service,
)
from semantic_digital_twin.datastructures.prefixed_name import PrefixedName
from semantic_digital_twin.semantic_annotations.semantic_annotations import Handle, Door
from semantic_digital_twin.testing import rclpy_node
from semantic_digital_twin.world import World
from semantic_digital_twin.world_description.connections import Connection6DoF
from semantic_digital_twin.world_description.world_entity import Body
from semantic_digital_twin.world_description.world_modification import (
    WorldModelModificationBlock,
)


def create_dummy_world():
    """
    Create a simple world with two bodies and a connection.
    """
    world = World()
    body_1 = Body(name=PrefixedName("body_1"))
    body_2 = Body(name=PrefixedName("body_2"))
    with world.modify_world():
        world.add_kinematic_structure_entity(body_1)
        world.add_kinematic_structure_entity(body_2)
        world.add_connection(
            Connection6DoF.create_with_dofs(parent=body_1, child=body_2, world=world)
        )
    return world


def test_get_modifications_as_json_empty_world(rclpy_node):
    """
    Test that get_modifications_as_json returns an empty list for a world with no modifications.
    """
    world = World()
    fetcher = FetchWorldServer(node=rclpy_node, world=world)

    modifications_json = fetcher.get_modifications_as_json()
    modifications_list = json.loads(modifications_json)

    assert modifications_list == []
    fetcher.close()


def test_service_callback_success(rclpy_node):
    """
    Test that the service callback returns success with the modifications JSON.
    """
    world = create_dummy_world()
    fetcher = FetchWorldServer(node=rclpy_node, world=world)

    # Create a mock request and response
    request = Trigger.Request()
    response = Trigger.Response()

    # Call the service callback directly
    result = fetcher.service_callback(request, response)

    assert result.success is True

    # Verify the message is valid JSON
    modifications_list = [
        WorldModelModificationBlock.from_json(d) for d in json.loads(result.message)
    ]

    assert modifications_list == world._model_modification_blocks

    fetcher.close()


def test_service_callback_with_multiple_modifications(rclpy_node):
    """
    Test that the service callback returns all modifications when multiple changes are made.
    """
    world = World()
    fetcher = FetchWorldServer(node=rclpy_node, world=world)

    # Make multiple modifications
    body_1 = Body(name=PrefixedName("body_1"))
    body_2 = Body(name=PrefixedName("body_2"))
    body_3 = Body(name=PrefixedName("body_3"))

    with world.modify_world():
        world.add_kinematic_structure_entity(body_1)

    with world.modify_world():
        world.add_kinematic_structure_entity(body_2)
        world.add_kinematic_structure_entity(body_3)
        world.add_connection(
            Connection6DoF.create_with_dofs(parent=body_1, child=body_2, world=world)
        )
        world.add_connection(
            Connection6DoF.create_with_dofs(parent=body_2, child=body_3, world=world)
        )

    request = Trigger.Request()
    response = Trigger.Response()

    result = fetcher.service_callback(request, response)

    assert result.success is True
    # Verify the message is valid JSON
    modifications_list = [
        WorldModelModificationBlock.from_json(d) for d in json.loads(result.message)
    ]
    assert modifications_list == world._model_modification_blocks
    fetcher.close()


def test_world_fetching(rclpy_node):
    world = create_dummy_world()
    fetcher = FetchWorldServer(node=rclpy_node, world=world)

    world2 = fetch_world_from_service(
        rclpy_node,
    )
    assert world2._model_modification_blocks == world._model_modification_blocks


def test_semantic_annotation_modifications(rclpy_node):
    w1 = World(name="w1")
    b1 = Body(name=PrefixedName("b1"))
    v1 = Handle(body=b1)
    v2 = Door(body=b1, handle=v1)

    with w1.modify_world():
        w1.add_body(b1)
        w1.add_semantic_annotation(v1)
        w1.add_semantic_annotation(v2)

    fetcher = FetchWorldServer(node=rclpy_node, world=w1)

    w2 = fetch_world_from_service(
        rclpy_node,
    )

    assert [sa.name for sa in w1.semantic_annotations] == [
        sa.name for sa in w2.semantic_annotations
    ]



# ----------------------------------------------------------------------------------------------------------------------
# This script generates the ORM classes for the semantic_digital_twin package.
# Dataclasses can be mapped automatically to the ORM model
# using the ORMatic library, they just have to be registered in the classes list.
# Classes that are self_mapped and explicitly_mapped are already mapped in the model.py file. Look there for more
# information on how to map them.
# ----------------------------------------------------------------------------------------------------------------------
from __future__ import annotations

import os
from dataclasses import is_dataclass

import krrood.entity_query_language.orm.model
import trimesh
from krrood.class_diagrams import ClassDiagram
from krrood.entity_query_language.predicate import HasTypes, HasType, Symbol
from krrood.entity_query_language.symbol_graph import SymbolGraph
from krrood.ormatic.dao import AlternativeMapping
from krrood.ormatic.ormatic import ORMatic
from krrood.ormatic.utils import classes_of_module
from krrood.utils import recursive_subclasses

import semantic_digital_twin.orm.model
import semantic_digital_twin.robots.abstract_robot
import semantic_digital_twin.semantic_annotations.semantic_annotations
import semantic_digital_twin.world  # ensure the module attribute exists on the package
import semantic_digital_twin.world_description.degree_of_freedom
import semantic_digital_twin.world_description.geometry
import semantic_digital_twin.world_description.shape_collection
import semantic_digital_twin.world_description.world_entity
from semantic_digital_twin.datastructures.prefixed_name import PrefixedName
from semantic_digital_twin.spatial_computations.forward_kinematics import (
    ForwardKinematicsVisitor,
)
from semantic_digital_twin.world import (
    ResetStateContextManager,
    WorldModelUpdateContextManager,
)
from semantic_digital_twin.world_description.connections import (
    FixedConnection,
    HasUpdateState,
)


# collect all semantic digital twin classes that should be mapped
all_classes = set(classes_of_module(semantic_digital_twin.orm.model))
all_classes |= set(
    classes_of_module(semantic_digital_twin.world_description.world_entity)
)
all_classes |= set(classes_of_module(semantic_digital_twin.world_description.geometry))
all_classes |= set(
    classes_of_module(semantic_digital_twin.world_description.shape_collection)
)
all_classes |= set(classes_of_module(semantic_digital_twin.world))
all_classes |= set(
    classes_of_module(semantic_digital_twin.datastructures.prefixed_name)
)

all_classes |= set(
    classes_of_module(semantic_digital_twin.world_description.connections)
)
all_classes |= set(
    classes_of_module(semantic_digital_twin.semantic_annotations.semantic_annotations)
)
all_classes |= set(
    classes_of_module(semantic_digital_twin.world_description.degree_of_freedom)
)
all_classes |= set(classes_of_module(semantic_digital_twin.robots.abstract_robot))


# remove classes that should not be mapped
all_classes -= {
    ResetStateContextManager,
    WorldModelUpdateContextManager,
    HasUpdateState,
    ForwardKinematicsVisitor,
}

# build the symbol graph
symbol_graph = SymbolGraph()

# collect all KRROOD classes
all_classes |= {c.clazz for c in symbol_graph.class_diagram.wrapped_classes}
all_classes |= {am.original_class() for am in recursive_subclasses(AlternativeMapping)}
all_classes |= set(classes_of_module(krrood.entity_query_language.symbol_graph))
all_classes |= {Symbol}

# remove classes that don't need persistence
all_classes -= {HasType, HasTypes}


# keep only dataclasses that are NOT AlternativeMapping subclasses
all_classes = {
    c for c in all_classes if is_dataclass(c) and not issubclass(c, AlternativeMapping)
}

# ensure we have the original classes of the mappings (ORMatic uses these)
all_classes |= {am.original_class() for am in recursive_subclasses(AlternativeMapping)}


def generate_orm():
    """
    Generate the ORM classes for the pycram package.
    """
    class_diagram = ClassDiagram(
        list(sorted(all_classes, key=lambda c: c.__name__, reverse=True))
    )

    instance = ORMatic(
        class_dependency_graph=class_diagram,
        type_mappings={trimesh.Trimesh: semantic_digital_twin.orm.model.TrimeshType},
        alternative_mappings=recursive_subclasses(AlternativeMapping),
    )

    instance.make_all_tables()

    script_dir = os.path.dirname(os.path.abspath(__file__))
    path = os.path.abspath(
        os.path.join(script_dir, "..", "src", "semantic_digital_twin", "orm")
    )
    with open(os.path.join(path, "ormatic_interface.py"), "w") as f:
        instance.to_sqlalchemy_file(f)


if __name__ == "__main__":
    generate_orm()



import os
import re
import unittest
from dataclasses import dataclass

import numpy as np

from semantic_digital_twin.adapters.fbx import FBXParser
from semantic_digital_twin.adapters.procthor.procthor_pipelines import (
    dresser_factory_from_body,
    drawer_factory_from_body,
    door_factory_from_body,
)
from semantic_digital_twin.datastructures.prefixed_name import PrefixedName
from semantic_digital_twin.pipeline.pipeline import (
    Step,
    Pipeline,
    BodyFilter,
    CenterLocalGeometryAndPreserveWorldPose,
    BodyFactoryReplace,
)
from semantic_digital_twin.spatial_types.spatial_types import TransformationMatrix
from semantic_digital_twin.utils import get_semantic_digital_twin_directory_root
from semantic_digital_twin.world import World
from semantic_digital_twin.world_description.connections import FixedConnection
from semantic_digital_twin.world_description.world_entity import Body


class PipelineTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.dummy_world = World()
        b1 = Body(name=PrefixedName("body1", "test"))
        b2 = Body(name=PrefixedName("body2", "test"))
        c1 = FixedConnection(b1, b2, TransformationMatrix())
        with cls.dummy_world.modify_world():
            cls.dummy_world.add_body(b1)
            cls.dummy_world.add_body(b2)
            cls.dummy_world.add_connection(c1)

        cls.fbx_path = os.path.join(
            get_semantic_digital_twin_directory_root(os.getcwd()),
            "resources",
            "fbx",
            "test_dressers.fbx",
        )

    def test_pipeline_and_step(self):

        @dataclass
        class TestStep(Step):
            body_name: PrefixedName

            def _apply(self, world: World) -> World:
                b1 = Body(name=self.body_name)
                world.add_body(b1)
                return world

        pipeline = Pipeline(steps=[TestStep(body_name=PrefixedName("body1", "test"))])

        dummy_world = World()

        dummy_world = pipeline.apply(dummy_world)

        self.assertEqual(len(dummy_world.bodies), 1)
        self.assertEqual(dummy_world.root.name, PrefixedName("body1", "test"))

    def test_body_filter(self):

        pipeline = Pipeline(
            steps=[BodyFilter(lambda x: x.name == PrefixedName("body1", "test"))]
        )

        filtered_world = pipeline.apply(self.dummy_world)
        self.assertEqual(len(filtered_world.bodies), 1)
        self.assertEqual(filtered_world.root.name, PrefixedName("body1", "test"))

    def test_center_local_geometry_and_preserve_world_pose(self):
        world = FBXParser(self.fbx_path).parse()

        original_bounding_boxes = [
            body.collision.as_bounding_box_collection_at_origin(
                TransformationMatrix(reference_frame=world.root)
            ).bounding_boxes[0]
            for body in world.bodies_with_enabled_collision
        ]

        original_global_poses = [
            body.global_pose.to_np() for body in world.bodies_with_enabled_collision
        ]
        for pose in original_global_poses:
            np.testing.assert_almost_equal(pose, np.eye(4))

        pipeline = Pipeline(steps=[CenterLocalGeometryAndPreserveWorldPose()])

        centered_world = pipeline.apply(world)

        centered_global_poses = [
            body.global_pose.to_np()
            for body in centered_world.bodies_with_enabled_collision
        ]
        for original, centered in zip(original_global_poses, centered_global_poses):
            assert not np.allclose(original, centered)

        new_bounding_boxes = [
            body.collision.as_bounding_box_collection_at_origin(
                TransformationMatrix(reference_frame=centered_world.root)
            ).bounding_boxes[0]
            for body in centered_world.bodies_with_enabled_collision
        ]

        self.assertEqual(original_bounding_boxes, new_bounding_boxes)

    def test_body_factory_replace(self):
        dresser_pattern = re.compile(r"^.*dresser_(?!drawer\b).*$", re.IGNORECASE)
        world = FBXParser(self.fbx_path).parse()

        self.assertIsNotNone(world.get_body_by_name("dresser_205"))
        self.assertIsNotNone(world.get_body_by_name("dresser_217"))
        self.assertFalse(world.semantic_annotations)

        procthor_factory_replace_pipeline = Pipeline(
            [
                BodyFactoryReplace(
                    body_condition=lambda b: bool(
                        dresser_pattern.fullmatch(b.name.name)
                    )
                    and not (
                        "drawer" in b.name.name.lower() or "door" in b.name.name.lower()
                    ),
                    factory_creator=dresser_factory_from_body,
                )
            ]
        )

        replaced_world = procthor_factory_replace_pipeline.apply(world)

        self.assertRaises(KeyError, replaced_world.get_body_by_name, "dresser_205")
        self.assertRaises(KeyError, replaced_world.get_body_by_name, "dresser_217")
        self.assertTrue(replaced_world.semantic_annotations)
        self.assertIsNotNone(
            replaced_world.get_semantic_annotation_by_name("dresser_205")
        )
        self.assertIsNotNone(
            replaced_world.get_semantic_annotation_by_name("dresser_217")
        )

    def test_dresser_factory_from_body(self):
        world = FBXParser(self.fbx_path).parse()

        self.assertIsNotNone(dresser := world.get_body_by_name("dresser_205"))

        dresser_factory = dresser_factory_from_body(dresser)

        self.assertEqual(dresser_factory.name.name, "dresser_205")
        self.assertEqual(len(dresser_factory.drawers_factories), 4)
        self.assertEqual(len(dresser_factory.parent_T_drawers), 4)

    def test_drawer_factory_from_body(self):
        world = FBXParser(self.fbx_path).parse()

        self.assertIsNotNone(drawer := world.get_body_by_name("dresser_drawer_205_1"))

        drawer_factory = drawer_factory_from_body(drawer)

        self.assertEqual(drawer_factory.name.name, "dresser_drawer_205_1")
        self.assertIsNotNone(drawer_factory.handle_factory)
        self.assertIsNotNone(drawer_factory.container_factory)

    def test_door_factory_from_body(self):
        world = FBXParser(self.fbx_path).parse()

        self.assertIsNotNone(door := world.get_body_by_name("dresser_door_217_1"))

        door_factory = door_factory_from_body(door)

        self.assertEqual(door_factory.name.name, "dresser_door_217_1")
        self.assertIsNotNone(door_factory.handle_factory)
        self.assertIsNotNone(door_factory.scale)


if __name__ == "__main__":
    unittest.main()



# Generated by ORMatic

from __future__ import annotations
from sqlalchemy import (
    Column,
    ForeignKey,
    Integer,
    String,
    Float,
    Boolean,
    DateTime,
    Enum,
    JSON,
)
from sqlalchemy.orm import relationship, Mapped, mapped_column, DeclarativeBase

import builtins
import krrood.entity_query_language.orm.model
import krrood.entity_query_language.predicate
import krrood.entity_query_language.symbol_graph
import semantic_digital_twin.datastructures.prefixed_name
import semantic_digital_twin.orm.model
import semantic_digital_twin.robots.abstract_robot
import semantic_digital_twin.semantic_annotations.semantic_annotations
import semantic_digital_twin.world_description.connections
import semantic_digital_twin.world_description.geometry
import semantic_digital_twin.world_description.shape_collection
import semantic_digital_twin.world_description.world_entity
import typing
import typing_extensions


from krrood.ormatic.dao import DataAccessObject
from krrood.ormatic.custom_types import TypeType


class Base(DeclarativeBase):
    type_mappings = {}


class BoundingBoxDAO(
    Base, DataAccessObject[semantic_digital_twin.world_description.geometry.BoundingBox]
):

    __tablename__ = "BoundingBoxDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    min_x: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    min_y: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    min_z: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    max_x: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    max_y: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    max_z: Mapped[builtins.float] = mapped_column(use_existing_column=True)

    origin_id: Mapped[int] = mapped_column(
        ForeignKey("TransformationMatrixMappingDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    origin: Mapped[TransformationMatrixMappingDAO] = relationship(
        "TransformationMatrixMappingDAO",
        uselist=False,
        foreign_keys=[origin_id],
        post_update=True,
    )


class CollisionCheckingConfigDAO(
    Base,
    DataAccessObject[
        semantic_digital_twin.world_description.world_entity.CollisionCheckingConfig
    ],
):

    __tablename__ = "CollisionCheckingConfigDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    buffer_zone_distance: Mapped[typing.Optional[builtins.float]] = mapped_column(
        use_existing_column=True
    )
    violated_distance: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    disabled: Mapped[typing.Optional[builtins.bool]] = mapped_column(
        use_existing_column=True
    )
    max_avoided_bodies: Mapped[builtins.int] = mapped_column(use_existing_column=True)


class ColorDAO(
    Base, DataAccessObject[semantic_digital_twin.world_description.geometry.Color]
):

    __tablename__ = "ColorDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    R: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    G: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    B: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    A: Mapped[builtins.float] = mapped_column(use_existing_column=True)


class FieldOfViewDAO(
    Base, DataAccessObject[semantic_digital_twin.robots.abstract_robot.FieldOfView]
):

    __tablename__ = "FieldOfViewDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    vertical_angle: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    horizontal_angle: Mapped[builtins.float] = mapped_column(use_existing_column=True)


class HasDoorsDAO(
    Base,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.HasDoors
    ],
):

    __tablename__ = "HasDoorsDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    doors: Mapped[typing.List[DoorDAO]] = relationship(
        "DoorDAO", foreign_keys="[DoorDAO.hasdoorsdao_doors_id]", post_update=True
    )


class HasDrawersDAO(
    Base,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.HasDrawers
    ],
):

    __tablename__ = "HasDrawersDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    drawers: Mapped[typing.List[DrawerDAO]] = relationship(
        "DrawerDAO",
        foreign_keys="[DrawerDAO.hasdrawersdao_drawers_id]",
        post_update=True,
    )


class Point3MappingDAO(
    Base, DataAccessObject[semantic_digital_twin.orm.model.Point3Mapping]
):

    __tablename__ = "Point3MappingDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    x: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    y: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    z: Mapped[builtins.float] = mapped_column(use_existing_column=True)

    reference_frame_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("KinematicStructureEntityDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    reference_frame: Mapped[KinematicStructureEntityDAO] = relationship(
        "KinematicStructureEntityDAO",
        uselist=False,
        foreign_keys=[reference_frame_id],
        post_update=True,
    )


class PredicateClassRelationDAO(
    Base,
    DataAccessObject[krrood.entity_query_language.symbol_graph.PredicateClassRelation],
):

    __tablename__ = "PredicateClassRelationDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    inferred: Mapped[builtins.bool] = mapped_column(use_existing_column=True)

    source_id: Mapped[int] = mapped_column(
        ForeignKey("WrappedInstanceMappingDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    target_id: Mapped[int] = mapped_column(
        ForeignKey("WrappedInstanceMappingDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    predicate_id: Mapped[int] = mapped_column(
        ForeignKey("BinaryPredicateDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    symbolgraphmappingdao_predicate_relations_id: Mapped[
        typing.Optional[builtins.int]
    ] = mapped_column(
        ForeignKey("SymbolGraphMappingDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    source: Mapped[WrappedInstanceMappingDAO] = relationship(
        "WrappedInstanceMappingDAO",
        uselist=False,
        foreign_keys=[source_id],
        post_update=True,
    )
    target: Mapped[WrappedInstanceMappingDAO] = relationship(
        "WrappedInstanceMappingDAO",
        uselist=False,
        foreign_keys=[target_id],
        post_update=True,
    )
    predicate: Mapped[BinaryPredicateDAO] = relationship(
        "BinaryPredicateDAO",
        uselist=False,
        foreign_keys=[predicate_id],
        post_update=True,
    )


class QuaternionMappingDAO(
    Base, DataAccessObject[semantic_digital_twin.orm.model.QuaternionMapping]
):

    __tablename__ = "QuaternionMappingDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    x: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    y: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    z: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    w: Mapped[builtins.float] = mapped_column(use_existing_column=True)

    reference_frame_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("KinematicStructureEntityDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    reference_frame: Mapped[KinematicStructureEntityDAO] = relationship(
        "KinematicStructureEntityDAO",
        uselist=False,
        foreign_keys=[reference_frame_id],
        post_update=True,
    )


class RotationMatrixMappingDAO(
    Base, DataAccessObject[semantic_digital_twin.orm.model.RotationMatrixMapping]
):

    __tablename__ = "RotationMatrixMappingDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    rotation_id: Mapped[int] = mapped_column(
        ForeignKey("QuaternionMappingDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    reference_frame_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("KinematicStructureEntityDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    rotation: Mapped[QuaternionMappingDAO] = relationship(
        "QuaternionMappingDAO",
        uselist=False,
        foreign_keys=[rotation_id],
        post_update=True,
    )
    reference_frame: Mapped[KinematicStructureEntityDAO] = relationship(
        "KinematicStructureEntityDAO",
        uselist=False,
        foreign_keys=[reference_frame_id],
        post_update=True,
    )


class ScaleDAO(
    Base, DataAccessObject[semantic_digital_twin.world_description.geometry.Scale]
):

    __tablename__ = "ScaleDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    x: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    y: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    z: Mapped[builtins.float] = mapped_column(use_existing_column=True)


class ShapeDAO(
    Base, DataAccessObject[semantic_digital_twin.world_description.geometry.Shape]
):

    __tablename__ = "ShapeDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    polymorphic_type: Mapped[str] = mapped_column(
        String(255), nullable=False, use_existing_column=True
    )

    origin_id: Mapped[int] = mapped_column(
        ForeignKey("TransformationMatrixMappingDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    color_id: Mapped[int] = mapped_column(
        ForeignKey("ColorDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    shapecollectiondao_shapes_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("ShapeCollectionDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    origin: Mapped[TransformationMatrixMappingDAO] = relationship(
        "TransformationMatrixMappingDAO",
        uselist=False,
        foreign_keys=[origin_id],
        post_update=True,
    )
    color: Mapped[ColorDAO] = relationship(
        "ColorDAO", uselist=False, foreign_keys=[color_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_on": "polymorphic_type",
        "polymorphic_identity": "ShapeDAO",
    }


class BoxDAO(
    ShapeDAO, DataAccessObject[semantic_digital_twin.world_description.geometry.Box]
):

    __tablename__ = "BoxDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(ShapeDAO.database_id), primary_key=True, use_existing_column=True
    )

    scale_id: Mapped[int] = mapped_column(
        ForeignKey("ScaleDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    scale: Mapped[ScaleDAO] = relationship(
        "ScaleDAO", uselist=False, foreign_keys=[scale_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "BoxDAO",
        "inherit_condition": database_id == ShapeDAO.database_id,
    }


class CylinderDAO(
    ShapeDAO,
    DataAccessObject[semantic_digital_twin.world_description.geometry.Cylinder],
):

    __tablename__ = "CylinderDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(ShapeDAO.database_id), primary_key=True, use_existing_column=True
    )

    width: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    height: Mapped[builtins.float] = mapped_column(use_existing_column=True)

    __mapper_args__ = {
        "polymorphic_identity": "CylinderDAO",
        "inherit_condition": database_id == ShapeDAO.database_id,
    }


class MeshDAO(
    ShapeDAO, DataAccessObject[semantic_digital_twin.world_description.geometry.Mesh]
):

    __tablename__ = "MeshDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(ShapeDAO.database_id), primary_key=True, use_existing_column=True
    )

    scale_id: Mapped[int] = mapped_column(
        ForeignKey("ScaleDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    scale: Mapped[ScaleDAO] = relationship(
        "ScaleDAO", uselist=False, foreign_keys=[scale_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "MeshDAO",
        "inherit_condition": database_id == ShapeDAO.database_id,
    }


class FileMeshDAO(
    MeshDAO, DataAccessObject[semantic_digital_twin.world_description.geometry.FileMesh]
):

    __tablename__ = "FileMeshDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(MeshDAO.database_id), primary_key=True, use_existing_column=True
    )

    filename: Mapped[builtins.str] = mapped_column(use_existing_column=True)

    __mapper_args__ = {
        "polymorphic_identity": "FileMeshDAO",
        "inherit_condition": database_id == MeshDAO.database_id,
    }


class ShapeCollectionDAO(
    Base,
    DataAccessObject[
        semantic_digital_twin.world_description.shape_collection.ShapeCollection
    ],
):

    __tablename__ = "ShapeCollectionDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    polymorphic_type: Mapped[str] = mapped_column(
        String(255), nullable=False, use_existing_column=True
    )

    reference_frame_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("KinematicStructureEntityDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    shapes: Mapped[typing.List[ShapeDAO]] = relationship(
        "ShapeDAO",
        foreign_keys="[ShapeDAO.shapecollectiondao_shapes_id]",
        post_update=True,
    )
    reference_frame: Mapped[KinematicStructureEntityDAO] = relationship(
        "KinematicStructureEntityDAO",
        uselist=False,
        foreign_keys=[reference_frame_id],
        post_update=True,
    )

    __mapper_args__ = {
        "polymorphic_on": "polymorphic_type",
        "polymorphic_identity": "ShapeCollectionDAO",
    }


class BoundingBoxCollectionDAO(
    ShapeCollectionDAO,
    DataAccessObject[
        semantic_digital_twin.world_description.shape_collection.BoundingBoxCollection
    ],
):

    __tablename__ = "BoundingBoxCollectionDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(ShapeCollectionDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "BoundingBoxCollectionDAO",
        "inherit_condition": database_id == ShapeCollectionDAO.database_id,
    }


class SphereDAO(
    ShapeDAO, DataAccessObject[semantic_digital_twin.world_description.geometry.Sphere]
):

    __tablename__ = "SphereDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(ShapeDAO.database_id), primary_key=True, use_existing_column=True
    )

    radius: Mapped[builtins.float] = mapped_column(use_existing_column=True)

    __mapper_args__ = {
        "polymorphic_identity": "SphereDAO",
        "inherit_condition": database_id == ShapeDAO.database_id,
    }


class SymbolDAO(Base, DataAccessObject[krrood.entity_query_language.predicate.Symbol]):

    __tablename__ = "SymbolDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    polymorphic_type: Mapped[str] = mapped_column(
        String(255), nullable=False, use_existing_column=True
    )

    __mapper_args__ = {
        "polymorphic_on": "polymorphic_type",
        "polymorphic_identity": "SymbolDAO",
    }


class PredicateDAO(
    SymbolDAO, DataAccessObject[krrood.entity_query_language.predicate.Predicate]
):

    __tablename__ = "PredicateDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(SymbolDAO.database_id), primary_key=True, use_existing_column=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "PredicateDAO",
        "inherit_condition": database_id == SymbolDAO.database_id,
    }


class BinaryPredicateDAO(
    PredicateDAO,
    DataAccessObject[krrood.entity_query_language.predicate.BinaryPredicate],
):

    __tablename__ = "BinaryPredicateDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(PredicateDAO.database_id), primary_key=True, use_existing_column=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "BinaryPredicateDAO",
        "inherit_condition": database_id == PredicateDAO.database_id,
    }


class PrefixedNameDAO(
    SymbolDAO,
    DataAccessObject[semantic_digital_twin.datastructures.prefixed_name.PrefixedName],
):

    __tablename__ = "PrefixedNameDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(SymbolDAO.database_id), primary_key=True, use_existing_column=True
    )

    name: Mapped[builtins.str] = mapped_column(use_existing_column=True)
    prefix: Mapped[typing.Optional[builtins.str]] = mapped_column(
        use_existing_column=True
    )

    worldstatemappingdao_names_id: Mapped[typing.Optional[builtins.int]] = (
        mapped_column(
            ForeignKey("WorldStateMappingDAO.database_id", use_alter=True),
            nullable=True,
            use_existing_column=True,
        )
    )

    __mapper_args__ = {
        "polymorphic_identity": "PrefixedNameDAO",
        "inherit_condition": database_id == SymbolDAO.database_id,
    }


class SymbolGraphMappingDAO(
    Base, DataAccessObject[krrood.entity_query_language.orm.model.SymbolGraphMapping]
):

    __tablename__ = "SymbolGraphMappingDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    instances: Mapped[typing.List[WrappedInstanceMappingDAO]] = relationship(
        "WrappedInstanceMappingDAO",
        foreign_keys="[WrappedInstanceMappingDAO.symbolgraphmappingdao_instances_id]",
        post_update=True,
    )
    predicate_relations: Mapped[typing.List[PredicateClassRelationDAO]] = relationship(
        "PredicateClassRelationDAO",
        foreign_keys="[PredicateClassRelationDAO.symbolgraphmappingdao_predicate_relations_id]",
        post_update=True,
    )


class TransformationMatrixMappingDAO(
    Base, DataAccessObject[semantic_digital_twin.orm.model.TransformationMatrixMapping]
):

    __tablename__ = "TransformationMatrixMappingDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    position_id: Mapped[int] = mapped_column(
        ForeignKey("Point3MappingDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    rotation_id: Mapped[int] = mapped_column(
        ForeignKey("QuaternionMappingDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    reference_frame_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("KinematicStructureEntityDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    child_frame_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("KinematicStructureEntityDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    position: Mapped[Point3MappingDAO] = relationship(
        "Point3MappingDAO", uselist=False, foreign_keys=[position_id], post_update=True
    )
    rotation: Mapped[QuaternionMappingDAO] = relationship(
        "QuaternionMappingDAO",
        uselist=False,
        foreign_keys=[rotation_id],
        post_update=True,
    )
    reference_frame: Mapped[KinematicStructureEntityDAO] = relationship(
        "KinematicStructureEntityDAO",
        uselist=False,
        foreign_keys=[reference_frame_id],
        post_update=True,
    )
    child_frame: Mapped[KinematicStructureEntityDAO] = relationship(
        "KinematicStructureEntityDAO",
        uselist=False,
        foreign_keys=[child_frame_id],
        post_update=True,
    )


class TriangleMeshDAO(
    MeshDAO,
    DataAccessObject[semantic_digital_twin.world_description.geometry.TriangleMesh],
):

    __tablename__ = "TriangleMeshDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(MeshDAO.database_id), primary_key=True, use_existing_column=True
    )

    mesh: Mapped[typing.Optional[semantic_digital_twin.orm.model.TrimeshType]] = (
        mapped_column(
            semantic_digital_twin.orm.model.TrimeshType,
            nullable=True,
            use_existing_column=True,
        )
    )

    __mapper_args__ = {
        "polymorphic_identity": "TriangleMeshDAO",
        "inherit_condition": database_id == MeshDAO.database_id,
    }


class Vector3MappingDAO(
    Base, DataAccessObject[semantic_digital_twin.orm.model.Vector3Mapping]
):

    __tablename__ = "Vector3MappingDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    x: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    y: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    z: Mapped[builtins.float] = mapped_column(use_existing_column=True)

    reference_frame_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("KinematicStructureEntityDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    reference_frame: Mapped[KinematicStructureEntityDAO] = relationship(
        "KinematicStructureEntityDAO",
        uselist=False,
        foreign_keys=[reference_frame_id],
        post_update=True,
    )


class WorldMappingDAO(
    Base, DataAccessObject[semantic_digital_twin.orm.model.WorldMapping]
):

    __tablename__ = "WorldMappingDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    name: Mapped[typing.Optional[builtins.str]] = mapped_column(
        use_existing_column=True
    )

    state_id: Mapped[int] = mapped_column(
        ForeignKey("WorldStateMappingDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    kinematic_structure_entities: Mapped[typing.List[KinematicStructureEntityDAO]] = (
        relationship(
            "KinematicStructureEntityDAO",
            foreign_keys="[KinematicStructureEntityDAO.worldmappingdao_kinematic_structure_entities_id]",
            post_update=True,
        )
    )
    connections: Mapped[typing.List[ConnectionDAO]] = relationship(
        "ConnectionDAO",
        foreign_keys="[ConnectionDAO.worldmappingdao_connections_id]",
        post_update=True,
    )
    semantic_annotations: Mapped[typing.List[SemanticAnnotationDAO]] = relationship(
        "SemanticAnnotationDAO",
        foreign_keys="[SemanticAnnotationDAO.worldmappingdao_semantic_annotations_id]",
        post_update=True,
    )
    degrees_of_freedom: Mapped[typing.List[DegreeOfFreedomMappingDAO]] = relationship(
        "DegreeOfFreedomMappingDAO",
        foreign_keys="[DegreeOfFreedomMappingDAO.worldmappingdao_degrees_of_freedom_id]",
        post_update=True,
    )
    state: Mapped[WorldStateMappingDAO] = relationship(
        "WorldStateMappingDAO", uselist=False, foreign_keys=[state_id], post_update=True
    )


class WorldEntityDAO(
    SymbolDAO,
    DataAccessObject[semantic_digital_twin.world_description.world_entity.WorldEntity],
):

    __tablename__ = "WorldEntityDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(SymbolDAO.database_id), primary_key=True, use_existing_column=True
    )

    name_id: Mapped[int] = mapped_column(
        ForeignKey("PrefixedNameDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    name: Mapped[PrefixedNameDAO] = relationship(
        "PrefixedNameDAO", uselist=False, foreign_keys=[name_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "WorldEntityDAO",
        "inherit_condition": database_id == SymbolDAO.database_id,
    }


class ConnectionDAO(
    WorldEntityDAO,
    DataAccessObject[semantic_digital_twin.world_description.world_entity.Connection],
):

    __tablename__ = "ConnectionDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(WorldEntityDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    worldmappingdao_connections_id: Mapped[typing.Optional[builtins.int]] = (
        mapped_column(
            ForeignKey("WorldMappingDAO.database_id", use_alter=True),
            nullable=True,
            use_existing_column=True,
        )
    )
    parent_id: Mapped[int] = mapped_column(
        ForeignKey("KinematicStructureEntityDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    child_id: Mapped[int] = mapped_column(
        ForeignKey("KinematicStructureEntityDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    parent_T_connection_expression_id: Mapped[int] = mapped_column(
        ForeignKey("TransformationMatrixMappingDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    connection_T_child_expression_id: Mapped[int] = mapped_column(
        ForeignKey("TransformationMatrixMappingDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    parent: Mapped[KinematicStructureEntityDAO] = relationship(
        "KinematicStructureEntityDAO",
        uselist=False,
        foreign_keys=[parent_id],
        post_update=True,
    )
    child: Mapped[KinematicStructureEntityDAO] = relationship(
        "KinematicStructureEntityDAO",
        uselist=False,
        foreign_keys=[child_id],
        post_update=True,
    )
    parent_T_connection_expression: Mapped[TransformationMatrixMappingDAO] = (
        relationship(
            "TransformationMatrixMappingDAO",
            uselist=False,
            foreign_keys=[parent_T_connection_expression_id],
            post_update=True,
        )
    )
    connection_T_child_expression: Mapped[TransformationMatrixMappingDAO] = (
        relationship(
            "TransformationMatrixMappingDAO",
            uselist=False,
            foreign_keys=[connection_T_child_expression_id],
            post_update=True,
        )
    )

    __mapper_args__ = {
        "polymorphic_identity": "ConnectionDAO",
        "inherit_condition": database_id == WorldEntityDAO.database_id,
    }


class ActiveConnectionDAO(
    ConnectionDAO,
    DataAccessObject[
        semantic_digital_twin.world_description.connections.ActiveConnection
    ],
):

    __tablename__ = "ActiveConnectionDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(ConnectionDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    frozen_for_collision_avoidance: Mapped[builtins.bool] = mapped_column(
        use_existing_column=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "ActiveConnectionDAO",
        "inherit_condition": database_id == ConnectionDAO.database_id,
    }


class ActiveConnection1DOFDAO(
    ActiveConnectionDAO,
    DataAccessObject[
        semantic_digital_twin.world_description.connections.ActiveConnection1DOF
    ],
):

    __tablename__ = "ActiveConnection1DOFDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(ActiveConnectionDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    multiplier: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    offset: Mapped[builtins.float] = mapped_column(use_existing_column=True)

    axis_id: Mapped[int] = mapped_column(
        ForeignKey("Vector3MappingDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    dof_name_id: Mapped[int] = mapped_column(
        ForeignKey("PrefixedNameDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    axis: Mapped[Vector3MappingDAO] = relationship(
        "Vector3MappingDAO", uselist=False, foreign_keys=[axis_id], post_update=True
    )
    dof_name: Mapped[PrefixedNameDAO] = relationship(
        "PrefixedNameDAO", uselist=False, foreign_keys=[dof_name_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "ActiveConnection1DOFDAO",
        "inherit_condition": database_id == ActiveConnectionDAO.database_id,
    }


class PrismaticConnectionDAO(
    ActiveConnection1DOFDAO,
    DataAccessObject[
        semantic_digital_twin.world_description.connections.PrismaticConnection
    ],
):

    __tablename__ = "PrismaticConnectionDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(ActiveConnection1DOFDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "PrismaticConnectionDAO",
        "inherit_condition": database_id == ActiveConnection1DOFDAO.database_id,
    }


class RevoluteConnectionDAO(
    ActiveConnection1DOFDAO,
    DataAccessObject[
        semantic_digital_twin.world_description.connections.RevoluteConnection
    ],
):

    __tablename__ = "RevoluteConnectionDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(ActiveConnection1DOFDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "RevoluteConnectionDAO",
        "inherit_condition": database_id == ActiveConnection1DOFDAO.database_id,
    }


class FixedConnectionDAO(
    ConnectionDAO,
    DataAccessObject[
        semantic_digital_twin.world_description.connections.FixedConnection
    ],
):

    __tablename__ = "FixedConnectionDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(ConnectionDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "FixedConnectionDAO",
        "inherit_condition": database_id == ConnectionDAO.database_id,
    }


class PassiveConnectionDAO(
    ConnectionDAO,
    DataAccessObject[
        semantic_digital_twin.world_description.connections.PassiveConnection
    ],
):

    __tablename__ = "PassiveConnectionDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(ConnectionDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "PassiveConnectionDAO",
        "inherit_condition": database_id == ConnectionDAO.database_id,
    }


class Connection6DoFDAO(
    PassiveConnectionDAO,
    DataAccessObject[
        semantic_digital_twin.world_description.connections.Connection6DoF
    ],
):

    __tablename__ = "Connection6DoFDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(PassiveConnectionDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    x_name_id: Mapped[int] = mapped_column(
        ForeignKey("PrefixedNameDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    y_name_id: Mapped[int] = mapped_column(
        ForeignKey("PrefixedNameDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    z_name_id: Mapped[int] = mapped_column(
        ForeignKey("PrefixedNameDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    qx_name_id: Mapped[int] = mapped_column(
        ForeignKey("PrefixedNameDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    qy_name_id: Mapped[int] = mapped_column(
        ForeignKey("PrefixedNameDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    qz_name_id: Mapped[int] = mapped_column(
        ForeignKey("PrefixedNameDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    qw_name_id: Mapped[int] = mapped_column(
        ForeignKey("PrefixedNameDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    x_name: Mapped[PrefixedNameDAO] = relationship(
        "PrefixedNameDAO", uselist=False, foreign_keys=[x_name_id], post_update=True
    )
    y_name: Mapped[PrefixedNameDAO] = relationship(
        "PrefixedNameDAO", uselist=False, foreign_keys=[y_name_id], post_update=True
    )
    z_name: Mapped[PrefixedNameDAO] = relationship(
        "PrefixedNameDAO", uselist=False, foreign_keys=[z_name_id], post_update=True
    )
    qx_name: Mapped[PrefixedNameDAO] = relationship(
        "PrefixedNameDAO", uselist=False, foreign_keys=[qx_name_id], post_update=True
    )
    qy_name: Mapped[PrefixedNameDAO] = relationship(
        "PrefixedNameDAO", uselist=False, foreign_keys=[qy_name_id], post_update=True
    )
    qz_name: Mapped[PrefixedNameDAO] = relationship(
        "PrefixedNameDAO", uselist=False, foreign_keys=[qz_name_id], post_update=True
    )
    qw_name: Mapped[PrefixedNameDAO] = relationship(
        "PrefixedNameDAO", uselist=False, foreign_keys=[qw_name_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "Connection6DoFDAO",
        "inherit_condition": database_id == PassiveConnectionDAO.database_id,
    }


class OmniDriveDAO(
    PassiveConnectionDAO,
    DataAccessObject[semantic_digital_twin.world_description.connections.OmniDrive],
):

    __tablename__ = "OmniDriveDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(PassiveConnectionDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    frozen_for_collision_avoidance: Mapped[builtins.bool] = mapped_column(
        use_existing_column=True
    )

    x_name_id: Mapped[int] = mapped_column(
        ForeignKey("PrefixedNameDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    y_name_id: Mapped[int] = mapped_column(
        ForeignKey("PrefixedNameDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    roll_name_id: Mapped[int] = mapped_column(
        ForeignKey("PrefixedNameDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    pitch_name_id: Mapped[int] = mapped_column(
        ForeignKey("PrefixedNameDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    yaw_name_id: Mapped[int] = mapped_column(
        ForeignKey("PrefixedNameDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    x_velocity_name_id: Mapped[int] = mapped_column(
        ForeignKey("PrefixedNameDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    y_velocity_name_id: Mapped[int] = mapped_column(
        ForeignKey("PrefixedNameDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    x_name: Mapped[PrefixedNameDAO] = relationship(
        "PrefixedNameDAO", uselist=False, foreign_keys=[x_name_id], post_update=True
    )
    y_name: Mapped[PrefixedNameDAO] = relationship(
        "PrefixedNameDAO", uselist=False, foreign_keys=[y_name_id], post_update=True
    )
    roll_name: Mapped[PrefixedNameDAO] = relationship(
        "PrefixedNameDAO", uselist=False, foreign_keys=[roll_name_id], post_update=True
    )
    pitch_name: Mapped[PrefixedNameDAO] = relationship(
        "PrefixedNameDAO", uselist=False, foreign_keys=[pitch_name_id], post_update=True
    )
    yaw_name: Mapped[PrefixedNameDAO] = relationship(
        "PrefixedNameDAO", uselist=False, foreign_keys=[yaw_name_id], post_update=True
    )
    x_velocity_name: Mapped[PrefixedNameDAO] = relationship(
        "PrefixedNameDAO",
        uselist=False,
        foreign_keys=[x_velocity_name_id],
        post_update=True,
    )
    y_velocity_name: Mapped[PrefixedNameDAO] = relationship(
        "PrefixedNameDAO",
        uselist=False,
        foreign_keys=[y_velocity_name_id],
        post_update=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "OmniDriveDAO",
        "inherit_condition": database_id == PassiveConnectionDAO.database_id,
    }


class DegreeOfFreedomMappingDAO(
    Base, DataAccessObject[semantic_digital_twin.orm.model.DegreeOfFreedomMapping]
):

    __tablename__ = "DegreeOfFreedomMappingDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    lower_limits: Mapped[typing.List[builtins.float]] = mapped_column(
        JSON, nullable=False, use_existing_column=True
    )
    upper_limits: Mapped[typing.List[builtins.float]] = mapped_column(
        JSON, nullable=False, use_existing_column=True
    )

    worldmappingdao_degrees_of_freedom_id: Mapped[typing.Optional[builtins.int]] = (
        mapped_column(
            ForeignKey("WorldMappingDAO.database_id", use_alter=True),
            nullable=True,
            use_existing_column=True,
        )
    )
    name_id: Mapped[int] = mapped_column(
        ForeignKey("PrefixedNameDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    name: Mapped[PrefixedNameDAO] = relationship(
        "PrefixedNameDAO", uselist=False, foreign_keys=[name_id], post_update=True
    )


class KinematicStructureEntityDAO(
    WorldEntityDAO,
    DataAccessObject[
        semantic_digital_twin.world_description.world_entity.KinematicStructureEntity
    ],
):

    __tablename__ = "KinematicStructureEntityDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(WorldEntityDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    index: Mapped[typing.Optional[builtins.int]] = mapped_column(
        use_existing_column=True
    )

    worldmappingdao_kinematic_structure_entities_id: Mapped[
        typing.Optional[builtins.int]
    ] = mapped_column(
        ForeignKey("WorldMappingDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "KinematicStructureEntityDAO",
        "inherit_condition": database_id == WorldEntityDAO.database_id,
    }


class BodyDAO(
    KinematicStructureEntityDAO,
    DataAccessObject[semantic_digital_twin.world_description.world_entity.Body],
):

    __tablename__ = "BodyDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(KinematicStructureEntityDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    visual_id: Mapped[int] = mapped_column(
        ForeignKey("ShapeCollectionDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    collision_id: Mapped[int] = mapped_column(
        ForeignKey("ShapeCollectionDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    collision_config_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("CollisionCheckingConfigDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    temp_collision_config_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("CollisionCheckingConfigDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    visual: Mapped[ShapeCollectionDAO] = relationship(
        "ShapeCollectionDAO", uselist=False, foreign_keys=[visual_id], post_update=True
    )
    collision: Mapped[ShapeCollectionDAO] = relationship(
        "ShapeCollectionDAO",
        uselist=False,
        foreign_keys=[collision_id],
        post_update=True,
    )
    collision_config: Mapped[CollisionCheckingConfigDAO] = relationship(
        "CollisionCheckingConfigDAO",
        uselist=False,
        foreign_keys=[collision_config_id],
        post_update=True,
    )
    temp_collision_config: Mapped[CollisionCheckingConfigDAO] = relationship(
        "CollisionCheckingConfigDAO",
        uselist=False,
        foreign_keys=[temp_collision_config_id],
        post_update=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "BodyDAO",
        "inherit_condition": database_id == KinematicStructureEntityDAO.database_id,
    }


class RegionDAO(
    KinematicStructureEntityDAO,
    DataAccessObject[semantic_digital_twin.world_description.world_entity.Region],
):

    __tablename__ = "RegionDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(KinematicStructureEntityDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    area_id: Mapped[int] = mapped_column(
        ForeignKey("ShapeCollectionDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    area: Mapped[ShapeCollectionDAO] = relationship(
        "ShapeCollectionDAO", uselist=False, foreign_keys=[area_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "RegionDAO",
        "inherit_condition": database_id == KinematicStructureEntityDAO.database_id,
    }


class SemanticAnnotationDAO(
    WorldEntityDAO,
    DataAccessObject[
        semantic_digital_twin.world_description.world_entity.SemanticAnnotation
    ],
):

    __tablename__ = "SemanticAnnotationDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(WorldEntityDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    worldmappingdao_semantic_annotations_id: Mapped[typing.Optional[builtins.int]] = (
        mapped_column(
            ForeignKey("WorldMappingDAO.database_id", use_alter=True),
            nullable=True,
            use_existing_column=True,
        )
    )

    __mapper_args__ = {
        "polymorphic_identity": "SemanticAnnotationDAO",
        "inherit_condition": database_id == WorldEntityDAO.database_id,
    }


class ComponentsDAO(
    SemanticAnnotationDAO,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.Components
    ],
):

    __tablename__ = "ComponentsDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(SemanticAnnotationDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "ComponentsDAO",
        "inherit_condition": database_id == SemanticAnnotationDAO.database_id,
    }


class DrawerDAO(
    ComponentsDAO,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.Drawer
    ],
):

    __tablename__ = "DrawerDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(ComponentsDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    hasdrawersdao_drawers_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("HasDrawersDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    container_id: Mapped[int] = mapped_column(
        ForeignKey("ContainerDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    handle_id: Mapped[int] = mapped_column(
        ForeignKey("HandleDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    cabinetdao_drawers_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("CabinetDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    dresserdao_drawers_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("DresserDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    wardrobedao_drawers_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("WardrobeDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    container: Mapped[ContainerDAO] = relationship(
        "ContainerDAO", uselist=False, foreign_keys=[container_id], post_update=True
    )
    handle: Mapped[HandleDAO] = relationship(
        "HandleDAO", uselist=False, foreign_keys=[handle_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "DrawerDAO",
        "inherit_condition": database_id == ComponentsDAO.database_id,
    }


class EntryWayDAO(
    ComponentsDAO,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.EntryWay
    ],
):

    __tablename__ = "EntryWayDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(ComponentsDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    body_id: Mapped[int] = mapped_column(
        ForeignKey("BodyDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    body: Mapped[BodyDAO] = relationship(
        "BodyDAO", uselist=False, foreign_keys=[body_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "EntryWayDAO",
        "inherit_condition": database_id == ComponentsDAO.database_id,
    }


class DoorDAO(
    EntryWayDAO,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.Door
    ],
):

    __tablename__ = "DoorDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(EntryWayDAO.database_id), primary_key=True, use_existing_column=True
    )

    hasdoorsdao_doors_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("HasDoorsDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    handle_id: Mapped[int] = mapped_column(
        ForeignKey("HandleDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    cabinetdao_doors_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("CabinetDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    cupboarddao_doors_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("CupboardDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    dresserdao_doors_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("DresserDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    wardrobedao_doors_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("WardrobeDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    walldao_doors_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("WallDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    handle: Mapped[HandleDAO] = relationship(
        "HandleDAO", uselist=False, foreign_keys=[handle_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "DoorDAO",
        "inherit_condition": database_id == EntryWayDAO.database_id,
    }


class DoubleDoorDAO(
    EntryWayDAO,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.DoubleDoor
    ],
):

    __tablename__ = "DoubleDoorDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(EntryWayDAO.database_id), primary_key=True, use_existing_column=True
    )

    left_door_id: Mapped[int] = mapped_column(
        ForeignKey("DoorDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    right_door_id: Mapped[int] = mapped_column(
        ForeignKey("DoorDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    left_door: Mapped[DoorDAO] = relationship(
        "DoorDAO", uselist=False, foreign_keys=[left_door_id], post_update=True
    )
    right_door: Mapped[DoorDAO] = relationship(
        "DoorDAO", uselist=False, foreign_keys=[right_door_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "DoubleDoorDAO",
        "inherit_condition": database_id == EntryWayDAO.database_id,
    }


class ContainerDAO(
    SemanticAnnotationDAO,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.Container
    ],
):

    __tablename__ = "ContainerDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(SemanticAnnotationDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    body_id: Mapped[int] = mapped_column(
        ForeignKey("BodyDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    body: Mapped[BodyDAO] = relationship(
        "BodyDAO", uselist=False, foreign_keys=[body_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "ContainerDAO",
        "inherit_condition": database_id == SemanticAnnotationDAO.database_id,
    }


class FridgeDAO(
    SemanticAnnotationDAO,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.Fridge
    ],
):

    __tablename__ = "FridgeDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(SemanticAnnotationDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    body_id: Mapped[int] = mapped_column(
        ForeignKey("BodyDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    door_id: Mapped[int] = mapped_column(
        ForeignKey("DoorDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    body: Mapped[BodyDAO] = relationship(
        "BodyDAO", uselist=False, foreign_keys=[body_id], post_update=True
    )
    door: Mapped[DoorDAO] = relationship(
        "DoorDAO", uselist=False, foreign_keys=[door_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "FridgeDAO",
        "inherit_condition": database_id == SemanticAnnotationDAO.database_id,
    }


class FurnitureDAO(
    SemanticAnnotationDAO,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.Furniture
    ],
):

    __tablename__ = "FurnitureDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(SemanticAnnotationDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "FurnitureDAO",
        "inherit_condition": database_id == SemanticAnnotationDAO.database_id,
    }


class CabinetDAO(
    FurnitureDAO,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.Cabinet
    ],
):

    __tablename__ = "CabinetDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(FurnitureDAO.database_id), primary_key=True, use_existing_column=True
    )

    container_id: Mapped[int] = mapped_column(
        ForeignKey("ContainerDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    container: Mapped[ContainerDAO] = relationship(
        "ContainerDAO", uselist=False, foreign_keys=[container_id], post_update=True
    )
    drawers: Mapped[typing.List[DrawerDAO]] = relationship(
        "DrawerDAO", foreign_keys="[DrawerDAO.cabinetdao_drawers_id]", post_update=True
    )
    doors: Mapped[typing.List[DoorDAO]] = relationship(
        "DoorDAO", foreign_keys="[DoorDAO.cabinetdao_doors_id]", post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "CabinetDAO",
        "inherit_condition": database_id == FurnitureDAO.database_id,
    }


class CupboardDAO(
    FurnitureDAO,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.Cupboard
    ],
):

    __tablename__ = "CupboardDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(FurnitureDAO.database_id), primary_key=True, use_existing_column=True
    )

    container_id: Mapped[int] = mapped_column(
        ForeignKey("ContainerDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    container: Mapped[ContainerDAO] = relationship(
        "ContainerDAO", uselist=False, foreign_keys=[container_id], post_update=True
    )
    doors: Mapped[typing.List[DoorDAO]] = relationship(
        "DoorDAO", foreign_keys="[DoorDAO.cupboarddao_doors_id]", post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "CupboardDAO",
        "inherit_condition": database_id == FurnitureDAO.database_id,
    }


class DresserDAO(
    FurnitureDAO,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.Dresser
    ],
):

    __tablename__ = "DresserDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(FurnitureDAO.database_id), primary_key=True, use_existing_column=True
    )

    container_id: Mapped[int] = mapped_column(
        ForeignKey("ContainerDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    container: Mapped[ContainerDAO] = relationship(
        "ContainerDAO", uselist=False, foreign_keys=[container_id], post_update=True
    )
    drawers: Mapped[typing.List[DrawerDAO]] = relationship(
        "DrawerDAO", foreign_keys="[DrawerDAO.dresserdao_drawers_id]", post_update=True
    )
    doors: Mapped[typing.List[DoorDAO]] = relationship(
        "DoorDAO", foreign_keys="[DoorDAO.dresserdao_doors_id]", post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "DresserDAO",
        "inherit_condition": database_id == FurnitureDAO.database_id,
    }


class WardrobeDAO(
    FurnitureDAO,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.Wardrobe
    ],
):

    __tablename__ = "WardrobeDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(FurnitureDAO.database_id), primary_key=True, use_existing_column=True
    )

    container_id: Mapped[int] = mapped_column(
        ForeignKey("ContainerDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    container: Mapped[ContainerDAO] = relationship(
        "ContainerDAO", uselist=False, foreign_keys=[container_id], post_update=True
    )
    drawers: Mapped[typing.List[DrawerDAO]] = relationship(
        "DrawerDAO", foreign_keys="[DrawerDAO.wardrobedao_drawers_id]", post_update=True
    )
    doors: Mapped[typing.List[DoorDAO]] = relationship(
        "DoorDAO", foreign_keys="[DoorDAO.wardrobedao_doors_id]", post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "WardrobeDAO",
        "inherit_condition": database_id == FurnitureDAO.database_id,
    }


class HandleDAO(
    SemanticAnnotationDAO,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.Handle
    ],
):

    __tablename__ = "HandleDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(SemanticAnnotationDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    body_id: Mapped[int] = mapped_column(
        ForeignKey("BodyDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    body: Mapped[BodyDAO] = relationship(
        "BodyDAO", uselist=False, foreign_keys=[body_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "HandleDAO",
        "inherit_condition": database_id == SemanticAnnotationDAO.database_id,
    }


class RoomDAO(
    SemanticAnnotationDAO,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.Room
    ],
):

    __tablename__ = "RoomDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(SemanticAnnotationDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    floor_id: Mapped[int] = mapped_column(
        ForeignKey("FloorDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    floor: Mapped[FloorDAO] = relationship(
        "FloorDAO", uselist=False, foreign_keys=[floor_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "RoomDAO",
        "inherit_condition": database_id == SemanticAnnotationDAO.database_id,
    }


class RootedSemanticAnnotationDAO(
    SemanticAnnotationDAO,
    DataAccessObject[
        semantic_digital_twin.world_description.world_entity.RootedSemanticAnnotation
    ],
):

    __tablename__ = "RootedSemanticAnnotationDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(SemanticAnnotationDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    root_id: Mapped[int] = mapped_column(
        ForeignKey("BodyDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    root: Mapped[BodyDAO] = relationship(
        "BodyDAO", uselist=False, foreign_keys=[root_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "RootedSemanticAnnotationDAO",
        "inherit_condition": database_id == SemanticAnnotationDAO.database_id,
    }


class AbstractRobotDAO(
    RootedSemanticAnnotationDAO,
    DataAccessObject[semantic_digital_twin.robots.abstract_robot.AbstractRobot],
):

    __tablename__ = "AbstractRobotDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(RootedSemanticAnnotationDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    torso_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("TorsoDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    default_collision_config_id: Mapped[int] = mapped_column(
        ForeignKey("CollisionCheckingConfigDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    torso: Mapped[TorsoDAO] = relationship(
        "TorsoDAO", uselist=False, foreign_keys=[torso_id], post_update=True
    )
    manipulators: Mapped[typing.List[ManipulatorDAO]] = relationship(
        "ManipulatorDAO",
        foreign_keys="[ManipulatorDAO.abstractrobotdao_manipulators_id]",
        post_update=True,
    )
    sensors: Mapped[typing.List[SensorDAO]] = relationship(
        "SensorDAO",
        foreign_keys="[SensorDAO.abstractrobotdao_sensors_id]",
        post_update=True,
    )
    manipulator_chains: Mapped[typing.List[KinematicChainDAO]] = relationship(
        "KinematicChainDAO",
        foreign_keys="[KinematicChainDAO.abstractrobotdao_manipulator_chains_id]",
        post_update=True,
    )
    sensor_chains: Mapped[typing.List[KinematicChainDAO]] = relationship(
        "KinematicChainDAO",
        foreign_keys="[KinematicChainDAO.abstractrobotdao_sensor_chains_id]",
        post_update=True,
    )
    default_collision_config: Mapped[CollisionCheckingConfigDAO] = relationship(
        "CollisionCheckingConfigDAO",
        uselist=False,
        foreign_keys=[default_collision_config_id],
        post_update=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "AbstractRobotDAO",
        "inherit_condition": database_id == RootedSemanticAnnotationDAO.database_id,
    }


class SemanticEnvironmentAnnotationDAO(
    RootedSemanticAnnotationDAO,
    DataAccessObject[
        semantic_digital_twin.world_description.world_entity.SemanticEnvironmentAnnotation
    ],
):

    __tablename__ = "SemanticEnvironmentAnnotationDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(RootedSemanticAnnotationDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "SemanticEnvironmentAnnotationDAO",
        "inherit_condition": database_id == RootedSemanticAnnotationDAO.database_id,
    }


class SemanticRobotAnnotationDAO(
    RootedSemanticAnnotationDAO,
    DataAccessObject[
        semantic_digital_twin.robots.abstract_robot.SemanticRobotAnnotation
    ],
):

    __tablename__ = "SemanticRobotAnnotationDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(RootedSemanticAnnotationDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "SemanticRobotAnnotationDAO",
        "inherit_condition": database_id == RootedSemanticAnnotationDAO.database_id,
    }


class KinematicChainDAO(
    SemanticRobotAnnotationDAO,
    DataAccessObject[semantic_digital_twin.robots.abstract_robot.KinematicChain],
):

    __tablename__ = "KinematicChainDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(SemanticRobotAnnotationDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    abstractrobotdao_manipulator_chains_id: Mapped[typing.Optional[builtins.int]] = (
        mapped_column(
            ForeignKey("AbstractRobotDAO.database_id", use_alter=True),
            nullable=True,
            use_existing_column=True,
        )
    )
    abstractrobotdao_sensor_chains_id: Mapped[typing.Optional[builtins.int]] = (
        mapped_column(
            ForeignKey("AbstractRobotDAO.database_id", use_alter=True),
            nullable=True,
            use_existing_column=True,
        )
    )
    tip_id: Mapped[int] = mapped_column(
        ForeignKey("BodyDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    manipulator_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("ManipulatorDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    tip: Mapped[BodyDAO] = relationship(
        "BodyDAO", uselist=False, foreign_keys=[tip_id], post_update=True
    )
    manipulator: Mapped[ManipulatorDAO] = relationship(
        "ManipulatorDAO", uselist=False, foreign_keys=[manipulator_id], post_update=True
    )
    sensors: Mapped[typing.List[SensorDAO]] = relationship(
        "SensorDAO",
        foreign_keys="[SensorDAO.kinematicchaindao_sensors_id]",
        post_update=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "KinematicChainDAO",
        "inherit_condition": database_id == SemanticRobotAnnotationDAO.database_id,
    }


class ArmDAO(
    KinematicChainDAO, DataAccessObject[semantic_digital_twin.robots.abstract_robot.Arm]
):

    __tablename__ = "ArmDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(KinematicChainDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "ArmDAO",
        "inherit_condition": database_id == KinematicChainDAO.database_id,
    }


class FingerDAO(
    KinematicChainDAO,
    DataAccessObject[semantic_digital_twin.robots.abstract_robot.Finger],
):

    __tablename__ = "FingerDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(KinematicChainDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "FingerDAO",
        "inherit_condition": database_id == KinematicChainDAO.database_id,
    }


class NeckDAO(
    KinematicChainDAO,
    DataAccessObject[semantic_digital_twin.robots.abstract_robot.Neck],
):

    __tablename__ = "NeckDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(KinematicChainDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    pitch_body_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("BodyDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    yaw_body_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("BodyDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    pitch_body: Mapped[BodyDAO] = relationship(
        "BodyDAO", uselist=False, foreign_keys=[pitch_body_id], post_update=True
    )
    yaw_body: Mapped[BodyDAO] = relationship(
        "BodyDAO", uselist=False, foreign_keys=[yaw_body_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "NeckDAO",
        "inherit_condition": database_id == KinematicChainDAO.database_id,
    }


class TorsoDAO(
    KinematicChainDAO,
    DataAccessObject[semantic_digital_twin.robots.abstract_robot.Torso],
):

    __tablename__ = "TorsoDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(KinematicChainDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "TorsoDAO",
        "inherit_condition": database_id == KinematicChainDAO.database_id,
    }


class ManipulatorDAO(
    SemanticRobotAnnotationDAO,
    DataAccessObject[semantic_digital_twin.robots.abstract_robot.Manipulator],
):

    __tablename__ = "ManipulatorDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(SemanticRobotAnnotationDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    abstractrobotdao_manipulators_id: Mapped[typing.Optional[builtins.int]] = (
        mapped_column(
            ForeignKey("AbstractRobotDAO.database_id", use_alter=True),
            nullable=True,
            use_existing_column=True,
        )
    )
    tool_frame_id: Mapped[int] = mapped_column(
        ForeignKey("BodyDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    front_facing_orientation_id: Mapped[int] = mapped_column(
        ForeignKey("QuaternionMappingDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    front_facing_axis_id: Mapped[int] = mapped_column(
        ForeignKey("Vector3MappingDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    tool_frame: Mapped[BodyDAO] = relationship(
        "BodyDAO", uselist=False, foreign_keys=[tool_frame_id], post_update=True
    )
    front_facing_orientation: Mapped[QuaternionMappingDAO] = relationship(
        "QuaternionMappingDAO",
        uselist=False,
        foreign_keys=[front_facing_orientation_id],
        post_update=True,
    )
    front_facing_axis: Mapped[Vector3MappingDAO] = relationship(
        "Vector3MappingDAO",
        uselist=False,
        foreign_keys=[front_facing_axis_id],
        post_update=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "ManipulatorDAO",
        "inherit_condition": database_id == SemanticRobotAnnotationDAO.database_id,
    }


class ParallelGripperDAO(
    ManipulatorDAO,
    DataAccessObject[semantic_digital_twin.robots.abstract_robot.ParallelGripper],
):

    __tablename__ = "ParallelGripperDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(ManipulatorDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    finger_id: Mapped[int] = mapped_column(
        ForeignKey("FingerDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    thumb_id: Mapped[int] = mapped_column(
        ForeignKey("FingerDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    finger: Mapped[FingerDAO] = relationship(
        "FingerDAO", uselist=False, foreign_keys=[finger_id], post_update=True
    )
    thumb: Mapped[FingerDAO] = relationship(
        "FingerDAO", uselist=False, foreign_keys=[thumb_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "ParallelGripperDAO",
        "inherit_condition": database_id == ManipulatorDAO.database_id,
    }


class SensorDAO(
    SemanticRobotAnnotationDAO,
    DataAccessObject[semantic_digital_twin.robots.abstract_robot.Sensor],
):

    __tablename__ = "SensorDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(SemanticRobotAnnotationDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    abstractrobotdao_sensors_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("AbstractRobotDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    kinematicchaindao_sensors_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("KinematicChainDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "SensorDAO",
        "inherit_condition": database_id == SemanticRobotAnnotationDAO.database_id,
    }


class CameraDAO(
    SensorDAO, DataAccessObject[semantic_digital_twin.robots.abstract_robot.Camera]
):

    __tablename__ = "CameraDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(SensorDAO.database_id), primary_key=True, use_existing_column=True
    )

    minimal_height: Mapped[builtins.float] = mapped_column(use_existing_column=True)
    maximal_height: Mapped[builtins.float] = mapped_column(use_existing_column=True)

    forward_facing_axis_id: Mapped[int] = mapped_column(
        ForeignKey("Vector3MappingDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )
    field_of_view_id: Mapped[int] = mapped_column(
        ForeignKey("FieldOfViewDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    forward_facing_axis: Mapped[Vector3MappingDAO] = relationship(
        "Vector3MappingDAO",
        uselist=False,
        foreign_keys=[forward_facing_axis_id],
        post_update=True,
    )
    field_of_view: Mapped[FieldOfViewDAO] = relationship(
        "FieldOfViewDAO",
        uselist=False,
        foreign_keys=[field_of_view_id],
        post_update=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "CameraDAO",
        "inherit_condition": database_id == SensorDAO.database_id,
    }


class SupportingSurfaceDAO(
    SemanticAnnotationDAO,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.SupportingSurface
    ],
):

    __tablename__ = "SupportingSurfaceDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(SemanticAnnotationDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    region_id: Mapped[int] = mapped_column(
        ForeignKey("RegionDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    region: Mapped[RegionDAO] = relationship(
        "RegionDAO", uselist=False, foreign_keys=[region_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "SupportingSurfaceDAO",
        "inherit_condition": database_id == SemanticAnnotationDAO.database_id,
    }


class FloorDAO(
    SupportingSurfaceDAO,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.Floor
    ],
):

    __tablename__ = "FloorDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(SupportingSurfaceDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "FloorDAO",
        "inherit_condition": database_id == SupportingSurfaceDAO.database_id,
    }


class TableDAO(
    SemanticAnnotationDAO,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.Table
    ],
):

    __tablename__ = "TableDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(SemanticAnnotationDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    top_id: Mapped[int] = mapped_column(
        ForeignKey("BodyDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    top: Mapped[BodyDAO] = relationship(
        "BodyDAO", uselist=False, foreign_keys=[top_id], post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "TableDAO",
        "inherit_condition": database_id == SemanticAnnotationDAO.database_id,
    }


class WallDAO(
    SemanticAnnotationDAO,
    DataAccessObject[
        semantic_digital_twin.semantic_annotations.semantic_annotations.Wall
    ],
):

    __tablename__ = "WallDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        ForeignKey(SemanticAnnotationDAO.database_id),
        primary_key=True,
        use_existing_column=True,
    )

    body_id: Mapped[int] = mapped_column(
        ForeignKey("BodyDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    body: Mapped[BodyDAO] = relationship(
        "BodyDAO", uselist=False, foreign_keys=[body_id], post_update=True
    )
    doors: Mapped[typing.List[DoorDAO]] = relationship(
        "DoorDAO", foreign_keys="[DoorDAO.walldao_doors_id]", post_update=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "WallDAO",
        "inherit_condition": database_id == SemanticAnnotationDAO.database_id,
    }


class WorldStateMappingDAO(
    Base, DataAccessObject[semantic_digital_twin.orm.model.WorldStateMapping]
):

    __tablename__ = "WorldStateMappingDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    data: Mapped[typing.List[builtins.float]] = mapped_column(
        JSON, nullable=False, use_existing_column=True
    )

    names: Mapped[typing.List[PrefixedNameDAO]] = relationship(
        "PrefixedNameDAO",
        foreign_keys="[PrefixedNameDAO.worldstatemappingdao_names_id]",
        post_update=True,
    )


class WrappedInstanceMappingDAO(
    Base,
    DataAccessObject[krrood.entity_query_language.orm.model.WrappedInstanceMapping],
):

    __tablename__ = "WrappedInstanceMappingDAO"

    database_id: Mapped[builtins.int] = mapped_column(
        Integer, primary_key=True, use_existing_column=True
    )

    symbolgraphmappingdao_instances_id: Mapped[typing.Optional[builtins.int]] = (
        mapped_column(
            ForeignKey("SymbolGraphMappingDAO.database_id", use_alter=True),
            nullable=True,
            use_existing_column=True,
        )
    )
    instance_id: Mapped[typing.Optional[builtins.int]] = mapped_column(
        ForeignKey("SymbolDAO.database_id", use_alter=True),
        nullable=True,
        use_existing_column=True,
    )

    instance: Mapped[SymbolDAO] = relationship(
        "SymbolDAO", uselist=False, foreign_keys=[instance_id], post_update=True
    )



from typing_extensions import Annotated, Literal
from typing_extensions import Sequence
import numpy as np
import numpy.typing as npt

AnyMatrix4x4 = Annotated[Sequence[Sequence[float]], Literal[4, 4]]
NpMatrix4x4 = Annotated[npt.NDArray[np.float64], Literal[4, 4]]
